#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     This is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import os
import logging
import sys
import mercantile
from osgeo import ogr
from pySmartDL import SmartDL
from cpuinfo import get_cpu_info
import queue
import concurrent.futures
import threading
from osm_fieldwork.sqlite import DataFile

# Instantiate logger
log = logging.getLogger(__name__)


def dlthread(dest, mirrors, tiles):
    """Thread to handle downloads for Queue"""
    if len(tiles) == 0:
        # epdb.st()
        return
    # counter = -1

    # start = datetime.now()

    # totaltime = 0.0
    logging.info(
        "Downloading %d tiles in thread %d to %s"
        % (len(tiles), threading.get_ident(), dest)
    )
    for tile in tiles:
        filespec = f"{tile[2]}/{tile[1]}/{tile[0]}"
        for site in mirrors:
            if site["source"] != "topo":
                filespec += "." + site["suffix"]
            url = site["url"]
            remote = url % filespec
            print("Getting file from: %s" % remote)
            # Create the subdirectories as pySmartDL doesn't do it for us
            if os.path.isdir(dest) is False:
                tmp = ""
                paths = dest.split("/")
                for i in paths[1:]:
                    tmp += "/" + i
                    if os.path.isdir(tmp):
                        continue
                    else:
                        os.mkdir(tmp)
                        logging.debug("Made %s" % tmp)

        try:
            if site["source"] == "topo":
                filespec += "." + site["suffix"]
            outfile = dest + "/" + filespec
            if not os.path.exists(outfile):
                dl = SmartDL(remote, dest=outfile, connect_default_logger=False)
                dl.start()
            else:
                logging.debug("%s exists!" % (outfile))
        except:
            logging.error("Couldn't download from %r: %s" % (filespec, dl.get_errors()))


class BaseMapper(object):
    def __init__(self, filespec=None, base=None, source=None):
        """Create an mbtiles basemap for ODK Collect"""
        geom = ogr.Open(filespec)
        layer = geom.GetLayer()
        x_min, x_max, y_min, y_max = layer.GetExtent()
        layer.GetSpatialRef()

        self.bbox = self.makeBbox(layer)
        self.tiles = list()
        self.base = base
        # sources for imagery
        self.source = source
        self.sources = dict()

        # Bing hybrid imagery
        url = "http://ecn.t0.tiles.virtualearth.net/tiles/h%s.png?g=129&mkt=en&stl=H"
        source = {
            "name": "Bing Maps Hybrid",
            "url": url,
            "suffix": "png",
            "source": "bing",
        }
        self.sources["bing"] = source

        # ERSI imagery
        url = "http://services.arcgisonline.com/arcgis/rest/services/World_Imagery/MapServer/tile/%s"
        source = {
            "name": "ESRI World Imagery",
            "url": url,
            "suffix": "jpg",
            "source": "esri",
        }
        self.sources["esri"] = source

        # USGS Topographical map
        url = "https://basemap.nationalmap.gov/ArcGIS/rest/services/USGSTopo/MapServer/tile/%s"
        source = {
            "name": "USGS Topographic Map",
            "url": url,
            "suffix": "jpg",
            "source": "topo",
        }
        self.sources["topo"] = source

        # Google Hybrid
        url = "https://mt0.google.com/vt?lyrs=h&x={x}&s=&y={y}&z={z}"
        source = {
            "name": "Google Hybrid",
            "url": url,
            "suffix": "png",
            "source": "google",
        }
        self.sources["google"] = source

        # OpenArialMap
        url = "https://tiles.openaerialmap.org/63962d0d9f665400075759be/0/63962d0d9f665400075759bf/{z}/{x}/{y}"
        source = {
            "name": "Open Aerial Map",
            "url": url,
            "suffix": "png",
            "source": "oam",
        }
        self.sources["oam"] = source

    def getFormat(self):
        return self.sources[self.source]["suffix"]

    def getTiles(self, zoom=None):
        """Get a list of tiles for the specifed zoom level"""
        info = get_cpu_info()
        cores = info["count"]

        self.tiles = list(
            mercantile.tiles(
                self.bbox[0], self.bbox[1], self.bbox[2], self.bbox[3], zoom
            )
        )
        total = len(self.tiles)
        logging.info("%d tiles for zoom level %d" % (len(self.tiles), zoom))
        chunk = round(len(self.tiles) / cores)
        queue.Queue(maxsize=cores)
        logging.info("%d threads, %d tiles" % (cores, total))

        mirrors = [self.sources[self.source]]
        # epdb.st()
        if len(self.tiles) < chunk or chunk == 0:
            dlthread(self.base, mirrors, self.tiles)
        else:
            with concurrent.futures.ThreadPoolExecutor(max_workers=cores) as executor:
                block = 0
                while block <= len(self.tiles):
                    executor.submit(
                        dlthread, self.base, mirrors, self.tiles[block : block + chunk]
                    )
                    logging.debug("Dispatching Block %d:%d" % (block, block + chunk))
                    block += chunk
                executor.shutdown()
            # logging.info("Had %r errors downloading %d tiles for data for %r" % (self.errors, len(tiles), os.path.basename(self.base)))

        return len(self.tiles)

    def tileExists(self, tile=list()):
        """See if a map tile already exists"""
        filespec = f"{self.base}{tile[2]}/{tile[1]}/{tile[0]}.{self.sources[{self.source}]['suffix']}"
        if os.path.exists(filespec):
            logging.debug("%s exists" % filespec)
            return True
        else:
            logging.debug("%s doesn't exists" % filespec)
            return False

    def makeBbox(self, layer=None):
        """Make a bounding box from a layer"""
        # left, bottom, right, top
        # minX: %d, minY: %d, maxX: %d, maxY: %d" %(env[0],env[2],env[1],env[3])
        for feature in layer:
            bbox = list(feature.GetGeometryRef().GetEnvelope())
            bbox = (bbox[0], bbox[2], bbox[1], bbox[3])
            print(bbox)
        return bbox


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Create an mbtiles basemap for ODK Collect"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-b", "--boundary", help="The boundary for the area you want")
    parser.add_argument("-z", "--zooms", default="12-17", help="The Zoom levels")
    parser.add_argument("-o", "--outfile", help="Output file name")
    parser.add_argument("-d", "--outdir", help="Output directory name for tile cache")
    parser.add_argument(
        "-s",
        "--source",
        default="esri",
        choices=["esri", "bing", "topo", "google", "oam"],
        help="Imagery source",
    )
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        root.addHandler(ch)


# Get all the zoom levels we want
zooms = list()
if args.zooms:
    if args.zooms.find("-") > 0:
        start = int(args.zooms.split("-")[0])
        end = int(args.zooms.split("-")[1])
        x = range(start, end)
        for i in x:
            zooms.append(i)
    elif args.zooms.find(",") > 0:
        levels = args.zooms.split(",")
        for level in levels:
            zooms.append(int(level))
    else:
        zooms.append(int(args.zooms))

# Make a bounding box from the boundary file
if not args.boundary:
    logging.error("You need to specify a boundary file!")
    parser.print_help()
    quit()

if not args.outdir:
    base = "/var/www/html"
else:
    base = args.outdir
base = f"{base}/{args.source}tiles"

if args.source:
    basemap = BaseMapper(args.boundary, base, args.source)
else:
    logging.error("You need to specify a source!")
    parser.print_help()
    quit()

if args.outfile is None:
    logging.error("You need to specify an mbtiles or sqlitedb file!!")
    parser.print_help()
    quit()

outf = DataFile(args.outfile, basemap.getFormat())
suffix = os.path.splitext(args.outfile)[1]
if suffix == ".mbtiles":
    outf.addBounds(basemap.bbox)
for level in zooms:
    basemap.getTiles(level)
    if args.outfile:
        # Create output database and specify image format, png, jpg, or tif
        outf.writeTiles(basemap.tiles, base)
    else:
        logging.info("Only downloading tiles to %s!" % base)



#!/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

# The origins of this file was called odk_request.py, and gave me the
# idea to do a much enhanced version. Contributors to the original are:
# Author: hcwinsemius <h.c.winsemius@gmail.com>
# Author: Ivan Gayton <ivangayton@gmail.com>
# Author: Reetta Valimaki <reetta.valimaki8@gmail.com>


import logging
import sys
import os
import requests
from requests.auth import HTTPBasicAuth
import json
import zlib
from datetime import datetime
from base64 import b64encode
import segno
import zlib

# Set log level for urllib
log_level = os.getenv("LOG_LEVEL", default="INFO")
logging.getLogger("urllib3").setLevel(log_level)

log = logging.getLogger(__name__)


class OdkCentral(object):
    def __init__(self,
                 url: str = None,
                 user: str = None,
                 passwd: str = None
                 ):
        """A Class for accessing an ODK Central server via it's REST API"""
        if not url:
            url = os.getenv("ODK_CENTRAL_URL", default=None)
        self.url = url
        if not user:
            user = os.getenv("ODK_CENTRAL_USER", default=None)
        self.user = user
        if not passwd:
            passwd = os.getenv("ODK_CENTRAL_PASSWD", default=None)
        self.passwd = passwd
        verify = os.getenv("ODK_CENTRAL_SECURE", default=True)
        if type(verify) == str:
            self.verify = eval(verify)
        else:
            self.verify = verify
        # These are settings used by ODK Collect
        self.general = {
            "form_update_mode": "match_exactly",
            "autosend": "wifi_and_cellular",
        }
        # If there is a config file with authentication setting, use that
        # so we don't have to supply this all the time. This is only used
        # when odk_client is used, and no parameters are passed in.
        if not self.url:
            log.debug("Configuring ODKCentral from file .odkcentral")
            home = os.getenv("HOME")
            config = ".odkcentral"
            filespec = home + "/" + config
            if os.path.exists(filespec):
                file = open(filespec, "r")
                for line in file:
                    # Support embedded comments
                    if line[0] == "#":
                        continue
                    # Read the config file for authentication settings
                    tmp = line.split("=")
                    if tmp[0] == "url":
                        self.url = tmp[1].strip("\n")
                    if tmp[0] == "user":
                        self.user = tmp[1].strip("\n")
                    if tmp[0] == "passwd":
                        self.passwd = tmp[1].strip("\n")
            else:
                log.warning(f"Authentication settings missing from {filespec}")
        else:
            log.debug(f"ODKCentral configuration parsed: {self.url}")
        # Base URL for the REST API
        self.version = "v1"
        log.debug(f"Using {self.version} API")
        self.base = self.url + "/" + self.version + "/"

        # Authentication data
        self.auth = HTTPBasicAuth(self.user, self.passwd)

        # Use a persistant connect, better for multiple requests
        self.session = requests.Session()

        # These are just cached data from the queries
        self.projects = dict()
        self.users = None

    def authenticate(self,
                     url:str = None,
                     user:str = None,
                     passwd:str = None
                     ):
        """Setup authenticate to an ODK Central server"""
        if not self.url:
            self.url = url
        if not self.user:
            self.user = user
        if not self.passwd:
            self.passwd = passwd
        # Enable persistent connection, create a cookie for this session
        self.session.headers.update({"accept": "odkcentral"})

        # Connect to the server
        return self.session.get(self.url, auth=self.auth, verify=self.verify)

    def listProjects(self):
        """Fetch a list of projects from an ODK Central server, and
        store it as an indexed list."""
        log.info("Getting a list of projects from %s" % self.url)
        url = f'{self.base}projects'
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        projects = result.json()
        for project in projects:
            if isinstance(project, dict):
                if project.get("id") is not None:
                    self.projects[project['id']] = project
            else:
                log.info("No projects returned. Is this a first run?")
        return projects

    def createProject(self,
                      name: str
                      ):
        """Create a new project on an ODK Central server if it doesn't
        already exist"""
        log.debug(f"Checking if project named {name} exists already")
        exists = self.findProject(name)
        if exists:
            log.debug(f"Project named {name} already exists.")
            return exists
        else:
            url = f"{self.base}projects"
            log.debug(f"POSTing project {name} to {url} with verify={self.verify}")
            try:
                result = self.session.post(
                    url, auth=self.auth, json={"name": name}, verify=self.verify, timeout=4
                )
                result.raise_for_status()
            except requests.exceptions.RequestException as e:
                log.error(e)
                log.error("Failed to submit to ODKCentral")
            json_response = result.json()
            log.debug(f"Returned: {json_response}")
            # update the internal list of projects
            self.listProjects()
            return json_response

    def deleteProject(self,
                      project_id: int
                      ):
        """Delete a project on an ODK Central server"""
        url = f"{self.base}projects/{project_id}"
        self.session.delete(url, auth=self.auth, verify=self.verify)
        # update the internal list of projects
        self.listProjects()
        return self.findProject(project_id)

    def findProject(self,
                    project_id: int,
                    name:str = None
                    ):
        """Get the project data from Central"""
        if self.projects:
            if name:
                for key, value in self.projects.items():
                    if name == value["name"]:
                        return value
            if project_id:
                for key, value in self.projects.items():
                    if project_id == value["id"]:
                        return value
        return None

    def findAppUser(self,
                    user_id: int,
                    name: str = None
                    ):
        """Get the data for an app user"""
        if self.appusers:
            if name is not None:
                result = [d for d in self.appusers if d['displayName']==name]
                if result:
                    return result[0]
                else:
                    log.debug(f"No user found with name: {name}")
                    return None
            if user_id is not None:
                result = [d for d in self.appusers if d['id']==user_id]
                if result:
                    return result[0]
                else:
                    log.debug(f"No user found with id: {user_id}")
                    return None
        return None

    def listUsers(self):
        """Fetch a list of users on the ODK Central server"""
        log.info("Getting a list of users from %s" % self.url)
        url = self.base + "users"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        self.users = result.json()
        return self.users

    def dump(self):
        """Dump internal data structures, for debugging purposes only"""
        # print("URL: %s" % self.url)
        # print("User: %s" % self.user)
        # print("Passwd: %s" % self.passwd)
        print("REST URL: %s" % self.base)

        print("There are %d projects on this server" % len(self.projects))
        for id, data in self.projects.items():
            print("\t %s: %s" % (id, data["name"]))
        if self.users:
            print("There are %d users on this server" % len(self.users))
            for data in self.users:
                print("\t %s: %s" % (data["id"], data["email"]))
        else:
            print("There are no users on this server")


class OdkProject(OdkCentral):
    """Class to manipulate a project on an ODK Central server"""

    def __init__(self, url=None, user=None, passwd=None):
        super().__init__(url, user, passwd)
        self.forms = None
        self.submissions = None
        self.data = None
        self.appusers = None
        self.id = None

    def getData(self,
                keyword: str
                ):
        return self.data[keyword]

    def listForms(self,
                  xform: str
                  ):
        """Fetch a list of forms in a project on an ODK Central server."""
        url = f"{self.base}projects/{xform}/forms"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        self.forms = result.json()
        return self.forms

    def listAppUsers(self,
                     projectId: int
                     ):
        """Fetch a list of app users for a project from an ODK Central server."""
        url = f"{self.base}projects/{projectId}/app-users"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        self.appusers = result.json()
        return self.appusers

    def listAssignments(self,
                        projectId: int
                        ):
        """List the Role & Actor assignments for users on a project"""
        url = f"{self.base}projects/{projectId}/assignments"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        return result.json()

    def getDetails(self,
                   projectId: int
                   ):
        """Get all the details for a project on an ODK Central server"""
        url = f"{self.base}projects/{projectId}"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        self.data = result.json()
        return self.data

    def dump(self):
        """Dump internal data structures, for debugging purposes only"""
        super().dump()
        if self.forms:
            print("There are %d forms in this project" % len(self.forms))
            for data in self.forms:
                print(
                    "\t %s(%s): %s" % (data["xmlFormId"], data["version"], data["name"])
                )
        if self.data:
            print("Project ID: %s" % self.data["id"])
        print("There are %d submissions in this project" % len(self.submissions))
        for data in self.submissions:
            print("\t%s: %s" % (data["instanceId"], data["createdAt"]))
        print("There are %d app users in this project" % len(self.appusers))
        for data in self.appusers:
            print("\t%s: %s" % (data["id"], data["displayName"]))


class OdkForm(OdkCentral):
    """Class to manipulate a from on an ODK Central server"""

    def __init__(self,
                 url:str = None,
                 user: str = None,
                 passwd: str = None
                 ):
        super().__init__(url, user, passwd)
        self.name = None
        # Draft is for a form that isn't published yet
        self.draft = True
        # this is only populated if self.getDetails() is called first.
        self.data = dict()
        self.attach = list()
        self.publish = True
        self.media = list()
        self.xml = None
        self.submissions = list()
        self.appusers = dict()
        # self.xmlFormId = None
        # self.projectId = None

    def getName(self):
        """Extract the name from a form on an ODK Central server"""
        if "name" in self.data:
            return self.data["name"]
        else:
            log.warning("Execute OdkForm.getDetails() to get this data.")

    def getFormId(self):
        """Extract the xmlFormId from a form on an ODK Central server"""
        if "xmlFormId" in self.data:
            return self.data["xmlFormId"]
        else:
            log.warning("Execute OdkForm.getDetails() to get this data.")

    def getDetails(self,
                   projectId: int,
                   xform: str
                   ):
        """Get all the details for a form on an ODK Central server"""
        url = f"{self.base}projects/{projectId}/forms/{xform}"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        self.data = result.json()
        return result

    def listSubmissions(self,
                        projectId,
                        xform: str
                        ):
        """Fetch a list of submission instances for a given form."""
        url = f"{self.base}projects/{projectId}/forms/{xform}/submissions"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        self.submissions = result.json()
        return self.submissions

    def listAssignments(self,
                        projectId: int,
                        xform: str
                        ):
        """List the Role & Actor assignments for users on a project"""
        url = f"{self.base}projects/{projectId}/forms/{xform}/assignments"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        return result.json()

    def getSubmissions(self,
                       projectId: int,
                       xform: str,
                       submission_id: int,
                       disk: bool = False,
                       json: bool = True
                       ):
        """Fetch a CSV file of the submissions without media to a survey form."""
        headers = {"Content-Type": "application/json"}
        now = datetime.now()
        timestamp = f"{now.year}_{now.hour}_{now.minute}"
        
        if json:
            url = self.base + f"projects/{projectId}/forms/{xform}.svc/Submissions"
            filespec = f"/tmp/{xform}_{timestamp}.json"
        else:
            url = self.base + f"projects/{projectId}/forms/{xform}/submissions"
            filespec = f"/tmp/{xform}_{timestamp}.csv"

        if submission_id:
            url = url + f"('{submission_id}')"

        result = self.session.get(url, auth=self.auth, headers=headers, verify=self.verify)
        if result.status_code == 200:
            if disk:
                # id = self.forms[0]['xmlFormId']
                try:
                    file = open(filespec, "xb")
                    file.write(result.content)
                except FileExistsError:
                    file = open(filespec, "wb")
                    file.write(result.content)
                log.info("Wrote output file %s" % filespec)
                file.close()
            return result.content
        else:
            log.error(f'Submissions for {projectId}, Form {xform}' + "doesn't exist")
            return None

    def getSubmissionMedia(self,
                           projectId: int,
                           xform: str
                           ):
        """Fetch a ZIP file of the submissions with media to a survey form."""
        url = self.base + f"projects/{projectId}/forms/{xform}/submissions.csv.zip"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        return result

    def addMedia(self,
                 media: str,
                 filespec: str
                 ):
        """Add a data file to this form"""
        # FIXME: this also needs the data
        self.media[filespec] = media

    def addXMLForm(self,
                   projectId: int,
                   xmlFormId: int,
                   xform: str
                   ):
        """Add an XML file to this form"""
        self.xml = xform

    def listMedia(self,
                  projectId: int,
                  xform: str
                  ):
        """List all the attchements for this form"""
        if self.draft:
            url = f"{self.base}projects/{projectId}/forms/{xform}/draft/attachments"
        else:
            url = f"{self.base}projects/{projectId}/forms/{xform}/attachments"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        self.media = result.json()
        return self.media

    def uploadMedia(self,
                    projectId: int,
                    xform: str,
                    filespec: str
                    ):
        """Upload an attachement to the ODK Central server"""
        title = os.path.basename(os.path.splitext(filespec)[0])
        datafile = f"{title}.geojson"
        xid = xform.split('_')[2]
        url = f"{self.base}projects/{projectId}/forms/{xid}/draft"
        result = self.session.post(url, auth=self.auth, verify=self.verify)
        if result.status_code == 200:
            log.debug(f"Modified {title} to draft")
        else:
            status = eval(result._content)
            log.error(f"Couldn't modify {title} to draft: {status['message']}")

        url = f"{self.base}projects/{projectId}/forms/{xid}/draft/attachments/{datafile}"
        headers = {"Content-Type": "*/*"}
        file = open(filespec, "rb")
        media = file.read()
        file.close()
        result = self.session.post(
            url, auth=self.auth, data=media, headers=headers, verify=self.verify
        )
        if result.status_code == 200:
            log.debug(f"Uploaded {filespec} to Central")
        else:
            status = eval(result._content)
            log.error(f"Couldn't upload {filespec} to Central: {status['message']}")

        return result

    def getMedia(self,
                 projectId: int,
                 xform: str,
                 filename: str
                 ):
        """Fetch a specific attachment by filename from a submission to a form."""
        if self.draft:
            url = f"{self.base}projects/{projectId}/forms/{xform}/draft/attachments/{filename}"
        else:
            url = f"{self.base}projects/{projectId}/forms/{xform}/attachments/{filename}"
        result = self.session.get(url, auth=self.auth, verify=self.verify)
        if result.status_code == 200:
            log.debug(f"fetched {filename} from Central")
        else:
            status = eval(result._content)
            log.error(f"Couldn't fetch {filename} from Central: {status['message']}")
        self.media = result.content
        return self.media

    def createForm(self,
                   projectId: int,
                   xform: str,
                   filespec: str,
                   draft: bool = False
                   ):
        """Create a new form on an ODK Central server"""
        if draft is not None:
            self.draft = draft
        headers = {"Content-Type": "application/xml"}
        if self.draft:
            url = f"{self.base}projects/{projectId}/forms/{xform}/draft?ignoreWarnings=true&publish=false"
        else:
            url = f"{self.base}projects/{projectId}/forms?ignoreWarnings=true&publish=true"

        # Read the XML or XLS file
        file = open(filespec, "rb")
        xml = file.read()
        file.close()
        log.info("Read %d bytes from %s" % (len(xml), filespec))

        result = self.session.post(url, auth=self.auth,  data=xml, headers=headers, verify=self.verify)
        # epdb.st()
        # FIXME: should update self.forms with the new form
        if result.status_code != 200:
            if result.status_code == 409:
                log.error(f"{xform} already exists on Central")
            else:
                status = eval(result._content)
                log.error(f"Couldn't create {xform} on Central: {status['message']}")

        return result.status_code

    def deleteForm(self,
                   projectId: int,
                   xform: str
                   ):
        """Delete a form from an ODK Central server"""
        # FIXME: If your goal is to prevent it from showing up on survey clients like ODK Collect, consider
        # setting its state to closing or closed
        if self.draft:
            url = f"{self.base}projects/{projectId}/forms/{xform}/draft"
        else:
            url = f"{self.base}projects/{projectId}/forms/{xform}"
        print(url)
        result = self.session.delete(url, auth=self.auth, verify=self.verify)
        return result

    def publishForm(self,
                    projectId: int,
                    xform: str
                    ):
        """Publish a draft form. When creating a form that isn't a draft, it can get publised then"""
        version = now = datetime.now().strftime("%Y-%m-%dT%TZ")
        if xform.find("_") > 0:
            xid = xform.split('_')[2]
        else:
            xid = xform

        url = f"{self.base}projects/{projectId}/forms/{xid}/draft/publish?version={version}"
        result = self.session.post(url, auth=self.auth, verify=self.verify)
        if result.status_code != 200:
            status = eval(result._content)
            log.error(f"Couldn't publish {xform} on Central: {status['message']}")
        else:
            log.info(f"Published {xform} on Central.")
        return result.status_code

    def dump(self):
        """Dump internal data structures, for debugging purposes only"""
        # super().dump()
        entries = len(self.media)
        print("Form has %d attachements" % entries)
        for form in self.media:
            if "name" in form:
                print("Name: %s" % form["name"])


class OdkAppUser(OdkCentral):
    def __init__(self, url=None, user=None, passwd=None):
        """A Class for app user data"""
        super().__init__(url, user, passwd)
        self.user = None
        self.qrcode = None
        self.id = None

    def create(self,
               projectId: int,
               name: str
               ):
        """Create a new app-user for a form"""
        url = f"{self.base}projects/{projectId}/app-users"
        result = self.session.post(
            url, auth=self.auth, json={"displayName": name}, verify=self.verify
        )
        self.user = name
        return result

    def delete(self,
               projectId: int,
               userId: int
               ):
        """Create a new app-user for a form"""
        url = f"{self.base}projects/{projectId}/app-users/{userId}"
        result = self.session.delete(url, auth=self.auth, verify=self.verify)
        return result

    def updateRole(self,
                   projectId: int,
                   xform: str,
                   roleId: int = 2,
                   actorId: int = None
                   ):
        """Update the role of an app user for a form"""
        log.info("Update access to XForm %s for %s" % (xform, actorId))
        url = f"{self.base}projects/{projectId}/forms/{xform}/assignments/{roleId}/{actorId}"
        result = self.session.post(url, auth=self.auth, verify=self.verify)
        return result

    def grantAccess(self,
                    projectId: int,
                    roleId: int = 2,
                    userId: int = None,
                    xform: str = None,
                    actorId: int = None
                    ):
        """Grant access to an app user for a form"""
        url = f"{self.base}projects/{projectId}/forms/{xform}/assignments/{roleId}/{actorId}"
        result = self.session.post(url, auth=self.auth, verify=self.verify)
        return result

    def createQRCode(self,
                     project_id: int,
                     token: str,
                     name: str
                     ):
        """Get the QR Code for an app-user"""
        log.info(
            'Generating QR Code for app-user "%s" for project %s' % (name, project_id)
        )
        self.settings = {
            "general": {
                "server_url": f"{self.base}key/{token}/projects/{project_id}",
                "form_update_mode": "manual",
                "basemap_source": "MapBox",
                "autosend": "wifi_and_cellular",
            },
            "project": {"name": f"{name}"},
            "admin": {},
        }
        qr_data = b64encode(zlib.compress(json.dumps(self.settings).encode("utf-8")))
        self.qrcode = segno.make(qr_data, micro=False)
        self.qrcode.save(f"{name}.png", scale=5)
        return qr_data


# This following code is only for debugging purposes, since this is easier
# to use a debugger with instead of pytest.
if __name__ == '__main__':
    logging.basicConfig(
        level=log_level,
        format=(
            "%(asctime)s.%(msecs)03d [%(levelname)s] "
            "%(name)s | %(funcName)s:%(lineno)d | %(message)s"
        ),
        datefmt="%y-%m-%d %H:%M:%S",
        stream=sys.stdout,
    )

    # Gotta start somewhere...
    project = OdkProject()
    # Start the persistent HTTPS connection to the ODK Central server
    project.authenticate()
    # Get a list of all the projects on this ODK Central server
    project.listProjects()
    # List all the users on this ODK Central server
    project.listUsers()
    # List all the forms for this project. FIXME: don't hardcode the project ID
    project.listForms(4)
    # List all the app users for this project. FIXME: don't hardcode the project ID
    project.listAppUsers(4)
    # List all the submissions for this project. FIXME: don't hardcode the project ID ad form name
    # project.listSubmissions(4, "cemeteries")
    # project.getSubmission(4, "cemeteries")
    # Dump all the internal data
    project.dump()

    # Form management
    form = OdkForm()
    form.authenticate()
    x = form.getDetails(4, "cemeteries")
    # print(x.json())
    # x = form.listMedia(4, "waterpoints", 'uuid:fbe3ef41-6298-40c1-a694-6c9d25a8c476')
    # Make a new form
    # xml = "/home/rob/projects/HOT/osm_fieldwork.git/osm_fieldwork/xlsforms/cemeteries.xml"
    # form.addXMLForm(xml)
    # csv1 = "/home/rob/projects/HOT/osm_fieldwork.git/osm_fieldwork/xlsforms/municipality.csv"
    # csv2 = "/home/rob/projects/HOT/osm_fieldwork.git/osm_fieldwork/xlsforms/towns.csv"
    # form.addMedia(csv1)
    # form.addMedia(csv2)
    x = form.createForm(4, "cemeteries", "cemeteries.xls", True)
    print(x.json())
    # x = form.publish(4, 'cemeteries', "cemeteries.xls")
    print(x.json())
    x = form.uploadMedia(4, "cemeteries", "towns.csv")
    print(x.json())
    x = form.uploadMedia(4, "cemeteries", "municipality.csv")
    print(x.json())
    x = form.listMedia(4, "cemeteries")
    print(x.json())
    form.dump()



#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import os
from osm_fieldwork.yamlfile import YamlFile
from osm_fieldwork.xlsforms import xlsforms_path
import logging
import argparse
import sys

# Instantiate logger
log = logging.getLogger(__name__)


class Convert(YamlFile):
    """A class to apply a YAML config file and convert ODK to OSM"""

    def __init__(self,
                 xform: str
                 ):
        path = xlsforms_path.replace("xlsforms", "")
        if type(xform) == str:
            file = f"{path}{xform}"
            xform = file
        else:
            file = f"{path}/xforms.yaml"
        self.yaml = YamlFile(file)
        self.filespec = file
        # Parse the file contents into a data structure to make it
        # easier to retrieve values
        self.convert = dict()
        self.ignore = list()
        self.private = list()
        for item in self.yaml.yaml["convert"]:
            key = list(item.keys())[0]
            value = item[key]
            # print("ZZZZ: %r, %r" % (key, value))
            if type(value) is str:
                self.convert[key] = value
            elif type(value) is list:
                vals = dict()
                for entry in value:
                    if type(entry) is str:
                        # epdb.st()
                        tag = entry
                    else:
                        tag = list(entry.keys())[0]
                        vals[tag] = entry[tag]
                self.convert[key] = vals
        self.ignore = self.yaml.yaml["ignore"]
        self.private = self.yaml.yaml["private"]
        self.multiple = self.yaml.yaml["multiple"]

    def privateData(self,
                    keyword: str
                    ):
        """See is a keyword is in the private data category"""
        return keyword in self.private

    def convertData(self,
                    keyword: str
                    ):
        """See is a keyword is in the convert data category"""
        return keyword in self.convert

    def ignoreData(self,
                   keyword: str
                   ):
        """See is a keyword is in the convert data category"""
        return keyword in self.ignore

    def escape(self,
               value: str
               ):
        """Escape characters like embedded quotes in text fields"""
        tmp = value.replace(" ", "_")
        return tmp.replace("'", "&apos;")

    def getKeyword(self,
                   value: str
                   ):
        """Get the value for a keyword from the yaml file"""
        key = self.yaml.yaml(value)
        if type(key) == bool:
            return value
        if len(key) == 0:
            key = self.yaml.getKeyword(value)
        return key

    def getValues(self,
                  tag: str = None
                  ):
        """Get the values for a primary key"""
        if tag is not None:
            if tag in self.convert:
                return self.convert[tag]
        else:
            return None

    def convertEntry(self,
                     tag: str,
                     value: str
                     ):
        """Convert a tag and value from the ODK represention to an OSM one"""
        #all = list()

        # If it's not in any conversion data, pass it through unchanged.
        if (
            tag not in self.convert
            and tag not in self.ignore
            and tag not in self.private
        ):
            return {tag: value}

        newtag = None
        newval = None
        # If the tag is in the config file, convert it.
        if self.convertData(tag):
            newtag = self.convertTag(tag)
            if newtag != tag:
                logging.debug(f"Converted Tag for entry {tag} to {newtag}")

        if newtag is None:
            newtag = tag
        # Truncate the elevation, as it's really long
        if newtag == "ele":
            value = value[:7]
        newval = self.convertValue(newtag, value)
        if newval != value:
            logging.debug("Converted Value for entry '%s' to '%s'" % (value, newval))
            for i in newval:
                key = list(i.keys())[0]
                newtag = key
                newval = i[key]
                #all.append({newtag: newval})
        #else:
        #    all.append({newtag: newval})

        # if newtag not in self.tags:
        #     tmp = { newtag: None }
        #     key = list()
        #     all.append(tmp)
        # else:
        #     val = self.tags[newtag]
        #     key = list(val.keys())[0]
        #     logging.debug("Converted Value for entry %s to %s" % (tag, value))
        #     # all = self.Value(tag, value)
        return {newtag: newval}

    def convertValue(self,
                     tag: str,
                     value: str
                     ):
        """Convert a single tag value"""
        all = list()

        vals = self.getValues(tag)
        # There is no conversion data for this tag
        if vals is None:
            return value

        if type(vals) is dict:
            if value not in vals:
                all.append({tag: value})
                return all
            if type(vals[value]) is bool:
                entry = dict()
                if vals[value]:
                    entry[tag] = "yes"
                else:
                    entry[tag] = "no"
                all.append(entry)
                return all
            for item in vals[value].split(","):
                entry = dict()
                tmp = item.split("=")
                if len(tmp) == 1:
                    entry[tag] = vals[value]
                else:
                    entry[tmp[0]] = tmp[1]
                    logging.debug("\tValue %s converted to %s" % (value, entry))
                all.append(entry)
        return all

    def convertTag(self,
                   tag: str
                   ):
        """Convert a single tag"""
        if tag in self.convert:
            newtag = self.convert[tag]
            if type(newtag) is str:
                logging.debug("\tTag '%s' converted to '%s'" % (tag, newtag))
                tmp = newtag.split("=")
                if len(tmp) > 1:
                    newtag = tmp[0]
            elif type(newtag) is list:
                logging.error("FIXME: list()")
                # epdb.st()
                return tag
            elif type(newtag) is dict:
                # logging.error("FIXME: dict()")
                return tag
            return newtag
        else:
            return tag

    def dump(self):
        """Dump the contents of the yaml file"""
        print("YAML file: %s" % self.filespec)
        for key, val in self.convert.items():
            if type(val) is list:
                print("Tag %s is" % key)
                for data in val:
                    print("\t%r" % data)
            else:
                print("Tag %s is %s" % (key, val))


#
# This script can be run standalone for debugging purposes. It's easier to debug
# this way than using pytest,
#
if __name__ == "__main__":
    # Enable logging to the terminal by default
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    root.addHandler(ch)

    parser = argparse.ArgumentParser(description="Read and parse a YAML file")
    parser.add_argument("-x", "--xform", default="xform.yaml", help="Default Yaml file")
    parser.add_argument(
        "-i",
        "--infile",
        help="The CSV input file",
    )
    args = parser.parse_args()

    # convert = Convert(args.xform)
    convert = Convert("xforms.yaml")
    print("-----")
    # tag = convert.convertTag("waterpoint_seasonal")
    # entry = convert.convertEntry("waterpoint_seasonal")
    # print("YY: %r" % entry)

    # print(convert.convertTag("tourism"))
    # entry = convert.convertEntry("tourism")
    # print(entry)
    # value = convert.convertEntry("waterpoint_seasonal")
    # print(value)
    # print("===============")

    # tag = convert.convertTag("waterpoint")
    # print(tag)
    # value = convert.convertValue("waterpoint", "well")
    # print(value)
    # value = convert.convertValue("power", "solar")

    entry = convert.convertEntry("waterpoint", "faucet")
    for i in entry:
        print("XX: %r" % i)

    entry = convert.convertEntry("operational_status", "closed")
    for i in entry:
        print("XX: %r" % i)

    entry = convert.convertEntry("seasonal", "wet")
    for i in entry:
        print("XX: %r" % i)

    entry = convert.convertEntry("seasonal", "rainy")
    for i in entry:
        print("XX: %r" % i)

    entry = convert.convertEntry("power", "solar")
    for i in entry:
        print("XX: %r" % i)



#!/usr/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of osm_fieldwork.
#
#     Underpass is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with osm_fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import sys
import os

sys.path.append(f"{os.getcwd()}/osm_fieldwork")
from yamlfile import YamlFile

parser = argparse.ArgumentParser(description="Read and parse a YAML file")
parser.add_argument(
    "--infile", default="osm_fieldwork/xforms.yaml", help="The YAML input file"
)
args = parser.parse_args()

data = YamlFile(args.infile)


def test_validate():
    """Tests for things under validate"""
    foo = data.yaml["validate"]
    bar = foo[1]["leisure"]
    assert "firepit" in bar


# def test_bool_good():
#     epdb.st()
#     assert "bear box" in data.yaml['convert'] is True

# def test_bool_bad():
#     assert "bad keyword" data.yaml['convert'] is not True


# def test_value():
#     assert "caravans" == "tourism"


if __name__ == "__main__":
    test_validate()
#    test_bool_good()
#    test_bool_bad()
#    test_value()



#!/usr/bin/python3

# Copyright (c) 2023 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import csv
import os
import logging
import sys
import json
import geojson
from sys import argv
from geojson import Point, Feature, FeatureCollection, dump
from pathlib import Path
import gpxpy
import gpxpy.gpx
from lxml import etree


# set log level for urlib
log = logging.getLogger(__name__)

def createExtension(icon):
    # camp_pitch.png, tourism_camp_site.png, topo_camp_pitch.png, topo_camp_site.png
    # trailhead.png, tourism_picnic_site.png, tourism_picnic_site.png,
    # tourism_attraction.png, tourism_information.png, information_board.png,
    # firepit.png, historic_ruins.png, amenity_drinking_water.png,
    # amenity_toilets.png, amenity_parking.png, special_trekking
    colors = {'tourism_camp_site': '#ff5020', 'tourism_picnic_site': '#ff5020', 'special_trekking': '#a71de1'}
    nsmap = {'osmand': 'https://osmand.net'}
    png = etree.Element('{osmand}icon', nsmap=nsmap)
    png.text = icon
    back = etree.Element('{osmand}background', nsmap=nsmap)
    back.text = "circle"
    color = None
    if icon in colors:
        color = etree.Element('{osmand}color', nsmap=nsmap)
        color.text = colors[icon]
        return (png, back, color)
    else:
        return (png, back)
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="convert GeoJson to a GPX favorites file for Osmand"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-i", "--infile", help="The data extract")
    args = parser.parse_args()

    if len(argv) <= 1:
        parser.print_help()
        quit()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        root.addHandler(ch)

        infile = open(args.infile, "r")
        indata = geojson.load(infile)

        # gpxpy.gpxfield.GPXField()
        gpx = gpxpy.gpx.GPX()
        gpx.nsmap['osmand'] = "https://osmand.net"
        gpx.creator = "osm2favorites 0.1"
        for feature in indata['features']:
            coords = feature['geometry']['coordinates']
            lat = coords[1]
            lon = coords[0]
            name = ""
            if 'name' in feature['properties']:
                name = feature['properties']['name']
                tourism = None
                if 'tourism' in feature['properties']:
                    tourism = feature['properties']['tourism']
                highway = None
                if 'highway' in feature['properties']:
                    highway = feature['properties']['highway']
                amenity = None
                if 'amenity' in feature['properties'] and not highway:
                    amenity = feature['properties']['amenity']
            for key, value in feature['properties'].items():
                if key == 'name':
                    continue
                description = "<p>"
                description += f"{key} = {value}<br>"
                description += "</p>"
            way = gpxpy.gpx.GPXWaypoint(
                latitude=lat,
                longitude=lon,
                description=description,
                name=name,
                # symbol="",
                # comment="",
            )
            if tourism and tourism != 'picnic site':
                extensions = createExtension("tourism_camp_site")
            elif tourism and tourism != 'picnic site':
                extensions = createExtension("tourism_picnic_site")
            elif highway and highway == 'trailhead':
                extensions = createExtension("special_trekking")
            elif amenity and amenity == 'parking':
                extensions = createExtension("amenity_parking")

            for ext in extensions:
                way.extensions.append(ext)
            gpx.waypoints.append(way)
        outfile = "output.gpx"
        with open(outfile, "w") as f:
            f.write(gpx.to_xml())

        log.info(f"Wrote {outfile}")



#!/bin/python3

# Copyright (c) 2022 Humanitarian OpenStreetMap Team
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import logging
import argparse
import sys
import os
import json
from osm_fieldwork.OdkCentral import OdkCentral, OdkProject, OdkForm, OdkAppUser
from pathlib import Path

# Set log level
log_level = os.getenv("LOG_LEVEL", default="INFO")
logging.getLogger("urllib3").setLevel(log_level)

log = logging.getLogger(__name__)


# FIXME: this classes isn;t used yet
class OdkClient(object):
    def __init__(self, url=None, user=None, passwd=None):
        """A Class for higher-level client side access to ODK Central"""
        self.url = url
        self.user = user
        self.passwd = passwd
        self.cache = dict()

    def readCache(self, cache=".odkcentral"):
        """FIXME: unimplemented"""
        file = open(cache, "rb")
        data = file.readline()
        print(json.load(data))
        file.close()

    def writeCache(self, cache=".odkcentral", data=None):
        """FIXME: unimplemented"""
        if args.cache:
            try:
                file = open(cache, "xb")
            except FileExistsError:
                file = open(cache, "wb")
            file.write("projects\n")
            file.write(json.dump(data))
            file.close()
        logging.info("Wrote config file %s" % filespec)


parser = argparse.ArgumentParser(description="command line client for ODK Central")
parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
# This is for server requests
parser.add_argument(
    "-s", "--server", choices=["projects", "users", "delete"], help="project operations"
)
# This is for project specific requests
parser.add_argument(
    "-p",
    "--project",
    choices=["forms", "app-users", "assignments", "delete"],
    help="project operations",
)
parser.add_argument("-i", "--id", type=int, help="Project ID nunmber")
parser.add_argument("-f", "--form", help="XForm name")
parser.add_argument("-u", "--uuid", help="Submission UUID, needed to download the data")
# This is for form specific requests
parser.add_argument("-x", "--xform",
    choices=[
        "attachments",
        "csv",
        "json",
        "submissions",
        "upload",
        "download",
        "create",
        "assignments",
        "delete",
        "publish",
    ],
    help="XForm ID for operations with data files",
)
parser.add_argument(
    "-a",
    "--appuser",
    choices=["create", "delete", "update", "qrcode", "access"],
    help="App-User operations",
)
parser.add_argument("-d", "--data", help="Data files for upload or download")
parser.add_argument("-t", "--timestamp", help="Timestamp for submissions")
parser.add_argument(
    "-b", "--bulk", choices=["qrcodes", "update"], help="Bulk operations"
)

# Creating a basic logger
if __name__ == '__main__':
    logging.basicConfig(
        level=log_level,
        format=(
            "%(asctime)s.%(msecs)03d [%(levelname)s]"
            "%(name)s | %(funcName)s:%(lineno)d | %(message)s"
        ),
        datefmt="%y-%m-%d %H:%M:%S",
        stream=sys.stdout,
    )


# Caching isn't implemented yet. That's for fancier queries that require multiple
# requests to the ODK server. Caching allows for data like names for IDs to
# be more user friendly.
# parser.add_argument('-c', '--cache', action="store_true", help = 'cache data from ODK Central')
# For now read these from the $HOME/.odkcentral config file
# parser.add_argument('-u', '--user', help = 'ODK Central username (usually email)')
# parser.add_argument('-pw', '--password', help = 'ODK Central password')

# parser.add_argument("-d", "--download", choices=['xml', 'xlsx', 'attach', 'submit', 'zip'], help="Download files from ODK Central")
# parser.add_argument("-u", "--upload", choices=['xml', 'xlsx', 'attach', 'submit', 'zip'], help="Upload files to ODK Central")

# Any files we want to use are specified on the command line with an argument.
# Multiple files are stored in a list.
args, unknown = parser.parse_known_args()

# Get any files for upload or download
files = list()
if unknown:
    files = unknown

# if verbose, dump to the terminal.
if args.verbose is not None:
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    root.addHandler(ch)

# Commands to the ODK Central server, which gets data that applies
# to all projects on the server.
if args.server:
    central = OdkCentral()
    # central.authenticate()
    if args.server == "projects":
        projects = central.listProjects()
        print("There are %d projects on this ODK Central server" % len(projects))
        ordered = sorted(projects, key=lambda item: item.get("id"))
        for project in ordered:
            print("\t%s: %s" % (project["id"], project["name"]))
    elif args.server == "users":
        users = central.listUsers()
        logging.info("There are %d users on this ODK Central server" % len(users))
        ordered = sorted(users, key=lambda item: item.get("id"))
        for user in ordered:
            print("%s: %s (%s)" % (user["id"], user["displayName"], user["email"]))
    elif args.server == "delete":
        central.deleteProject(args.id)

# Commands to get data about a specific project on an ODK Central server.
elif args.project:
    project = OdkProject()
    # project.authenticate()
    if not args.id:
        print('Need to specify a project ID using "--id"!')
        print(
            'You can use "odk_client.-py --server projects" to list all the projects!'
        )
        # parser.print_help()
        quit()
    if args.project == "forms":
        forms = project.listForms(args.id)
        ordered = sorted(forms, key=lambda item: item.get("xmlFormId"))
        for form in ordered:
            print("\t%r: %r" % (form["xmlFormId"], form["name"]))
    # if args.project == "submissions":
    #     submit = project.listSubmissions(args.id, args.form)
    #     # ordered = sorted(submit, key=lambda item: item.get('xmlFormId'))
    #     for data in submit:
    #         print("\t%s by user %s" % (data['instanceId'], data['submitterId']))
    if args.project == "app-users":
        users = project.listAppUsers(args.id)
        logging.info("There are %d app users on this ODK Central server" % len(users))
        ordered = sorted(users, key=lambda item: item.get("id"))
        for user in ordered:
            print("\t%r: %s (%s)" % (user["id"], user["displayName"], user["token"]))
    if args.project == "delete":
        tmp = files[0].split("-")
        if len(tmp) > 1:
            for id in range(int(tmp[0]), int(tmp[1])):
                project.deleteProject(id)
        else:
            project.deleteProject(tmp[0])

        # logging.info("There are %d app users on this ODK Central server" %)
    if args.project == "assignments":
        assign = project.listAssignments(args.id)
        logging.info(
            "There are %d assignments  on this ODK Central server" % len(assign)
        )
        ordered = sorted(assign, key=lambda item: item.get("id"))
        for role in ordered:
            print("\t%r" % role)

elif args.xform:
    # This downloads files from the ODK server
    if not args.id:
        print('Need to specify a project ID using "--id" and an XForm id using "--"!')
        quit()
    if not args.form:
        print('Need to specify a XForm id using "--form"!')
        quit()

    form = OdkForm()
    # form.authenticate()
    # Note that uploading and downloading is only for the attachments, usually
    # a CSV or GeoJson file used by the Form as an external data source for
    # survey questions
    if args.xform == "upload":
        for file in files:
            logging.info("Uploading file %r for XForm %s" % (file, args.form))
            result = form.uploadMedia(args.id, args.form, file)

    elif args.xform == "download":
        logging.info("Downloading files %r for XForm %s" % (files, args.form))
        for file in files:
            logging.info("Downloading %r for XForm %s" % (file, args.form))
            data = form.getMedia(args.id, args.form, file)
            try:
                file = open(file, "xb")
            except FileExistsError:
                file = open(file, "wb")
            file.write(data)
            file.close()
    elif args.xform == "assignments":
        assign = form.listAssignments(args.id, args.form)
        logging.info(
            "There are %d assignments  on this ODK Central server" % len(assign)
        )
        # ordered = sorted(assign, key=lambda item: item.get('id'))
        for role in assign:
            print("\t%r" % role)
    elif args.xform == "submissions":
        submissions = form.listSubmissions(args.id, args.form)
        logging.info(
            "There are %d submissions for XForm %s" % (len(submissions), args.form)
        )
        for file in submissions:
            # form.submissions.append(file)
            print("%s: %s" % (file["instanceId"], file["createdAt"]))

    elif args.xform == "csv":
        submissions = form.getSubmissions(args.id, args.form, True, False)
        logging.info(
            "There are %d submissions for XForm %s" % (len(submissions), args.form)
        )
        for file in submissions:
            form.submissions.append(file)
            print("%s: %s" % (file["instanceId"], file["createdAt"]))

    elif args.xform == "json":
        submissions = form.getSubmissions(args.id, args.form, False, True, True)
        logging.info(
            "There are %d submissions for XForm %s" % (len(submissions), args.form)
        )
        for file in submissions:
            form.submissions.append(file)
            #print("%s: %s" % (file["instanceId"], file["createdAt"]))

    elif args.xform == "attachments":
        attachments = form.listMedia(args.id, args.form)
        logging.info(
            "There are %d attachments for XForm %s" % (len(attachments), args.form)
        )
        for file in attachments:
            print("\t%s exists ? %s" % (file["name"], file["exists"]))

    elif args.xform == "create":
        for file in files:
            path = Path(file)
            if path.suffix == ".xml":
                logging.info("Creating XForm from %s" % file)
                result = form.createForm(args.id, args.form, file, True)
            elif path.suffix == ".csv":
                logging.info("Uploading media file %r for XForm %s" % (file, file))
                result = form.uploadMedia(args.id, args.form, file)
        result = form.publishForm(args.id, args.form)

    elif args.xform == "delete":
        logging.info("Deleting XForm from %s" % args.form)
        result = form.deleteForm(args.id, args.form)

    elif args.xform == "publish":
        logging.info("Publishing XForm from %s" % args.form)
        result = form.publishForm(args.id, args.form)

elif args.appuser:
    # This handles app-users
    print("App User ops %s" % args.appuser)
    if not args.id:
        print('Need to specify a project ID using "--id" and an XForm id using "--"!')
        quit()
    # if not args.form:
    #     print("Need to specify a XForm id using \"--form\"!")
    #     quit()

    role = 2  # seems to be the default value
    user = OdkAppUser()
    if args.appuser == "create":
        for appuser in files:
            result = user.create(args.id, appuser)
    elif args.appuser == "delete":
        tmp = files[0].split("-")
        if len(tmp) > 1:
            for id in range(int(tmp[0]), int(tmp[1])):
                result = user.delete(args.id, id)
        else:
            result = user.delete(args.id, tmp[0])
    elif args.appuser == "update":
        for appuser in files:
            result = user.updateRole(args.id, args.form, role, appuser)
    elif args.appuser == "qrcode":
        result = user.getQRCode(args.id, args.uuid, args.form)
    elif args.appuser == "access":
        for appuser in files:
            result = user.grantAccess(
                args.id,
                role,
                appuser,
            )
    print(result)

elif args.bulk:
    central = OdkProject()
    # project = central.getDetails(args.id)
    appuser = OdkAppUser()
    role = 2  # thise seems to be the default value
    if not args.form:
        print('Need to specify a XForm id using "--form"!')
        quit()
    if args.bulk == "qrcodes":
        users = central.listAppUsers(args.id)
        for user in users:
            # name = "%s-%s" % (project['name'], user['displayName'])
            name = "%s-%s" % (args.form, user["displayName"])
            result = appuser.getQRCode(args.id, user["token"], name)
    elif args.bulk == "update":
        users = central.listAppUsers(args.id)
        for user in users:
            result = appuser.updateRole(args.id, args.form, role, user["id"])



#!/usr/bin/python3

#
#   Copyright (C) 2023 Humanitarian OpenstreetMap Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import logging
import argparse
import xmltodict
import os
import sys
import re
from collections import OrderedDict
from pathlib import Path
from datetime import datetime
from osm_fieldwork.ODKInstance import ODKInstance
import geojson
import json
from geojson import Feature, FeatureCollection, dump
from shapely.geometry import Point
from datetime import datetime


# Instantiate logger
log = logging.getLogger(__name__)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert ODK XML instance file to GeoJson"
    )
    parser.add_argument("-v", "--verbose", nargs="?", const="0", help="verbose output")
    parser.add_argument(
        "-i", "--instance", help="The instance file(s) from ODK Collect"
    )
    parser.add_argument("-o","--outfile", default='tmp.geojson', help='The output file for JOSM')
    args = parser.parse_args()

    # if verbose, dump to the termina
    if not args.verbose:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        root.addHandler(ch)

    xmlfiles = list()
    if args.instance.find("*") >= 0:
        toplevel = Path()
        for dir in toplevel.glob(args.instance):
            if dir.is_dir():
                xml = os.listdir(dir)
                # There is always only one XML file per instance
                full = os.path.join(dir, xml[0])
                xmlfiles.append(full)
    else:
        toplevel = Path(args.instance)
        if toplevel.is_dir():
            # There is always only one XML file per instance
            full = os.path.join(toplevel, os.path.basename(toplevel))
            xmlfiles.append(full + ".xml")

    # print(xmlfiles)

    # These are all generated by Collect, and can be ignored
    rows = list()
    features = list()
    for xml in xmlfiles:
        print("Processing instance file: %s" % xml)
        tags = dict()
        #odkxml = ODKInstance()
        #data = odkxml.parse(filespec=xml)
        file = open(xml, "r")
        # Instances are small, read the whole file
        data = file.read(os.path.getsize(xml))
        doc = xmltodict.parse(data)
        # This is pretty ugly. We know that somewhere in the lower level
        # dicts there are GPS coordinates, which luckily are easy to
        # find and should always be a unique field in a feature. While
        # is ugly, it's also way more efficient than looping through
        # levels to find that one value. Plus htis doesn't care what the
        # key is.
        pat = re.compile("[0-9.]* [0-9.-]* [0-9.]* [0-9.]*")
        data = json.dumps(doc)
        gps = re.findall(pat, str(data))
        for coords in gps:
            tmp = coords.split(' ')
            lat = float(tmp[0])
            lon = float(tmp[1])                
            poi = Point(lon, lat)
        features.append(Feature(geometry=poi, properties=doc['data']))
    collection = FeatureCollection(features)

    now = datetime.now()
    timestamp = f"_{now.year}_{now.month}-{now.day}-{now.hour}-{now.minute}"
    outfile = args.instance.replace("*", "") + timestamp + ".geojson"
    json = open(outfile, 'w')
    dump(collection, json)

    print(f"Wrote output file {outfile}")



#!/usr/bin/python3
#
# Copyright (C) 2020, 2021, 2022, 2023 Humanitarian OpenstreetMap Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import logging
from datetime import datetime
from osm_fieldwork.convert import Convert
from osm_fieldwork.xlsforms import xlsforms_path
import xmltodict
import argparse
from sys import argv
import os
import pathlib

# Intantiate logger
log = logging.getLogger(__name__)


class OsmFile(object):
    """OSM File output"""

    def __init__(self,
                 filespec: str,
                 options: dict = dict(),
                 outdir: str = "/tmp/"
                 ):
        self.options = options
        # Read the config file to get our OSM credentials, if we have any
        # self.config = config.config(self.options)
        self.version = 3
        self.visible = "true"
        self.osmid = -1
        # Open the OSM output file
        if filespec is None:
            if "outdir" in self.options:
                self.file = self.options.get("outdir") + "foobar.osm"
        else:
            self.file = open(filespec, "w")
            # self.file = open(filespec + ".osm", 'w')
            logging.info("Opened output file: " + filespec)
        self.header()
        # logging.error("Couldn't open %s for writing!" % filespec)

        # This is the file that contains all the filtering data
        # self.ctable = convfile(self.options.get('convfile'))
        # self.options['convfile'] = None
        # These are for importing the CO addresses
        self.full = None
        self.addr = None
        # decrement the ID
        self.start = -1
        # path = xlsforms_path.replace("xlsforms", "")
        self.convert = Convert("xforms.yaml")
        self.data = dict()

    def isclosed(self):
        return self.file.closed

    def header(self):
        if self.file is not False:
            self.file.write("<?xml version='1.0' encoding='UTF-8'?>\n")
            # self.file.write('<osm version="0.6" generator="osm-fieldowrk 0.3" timestamp="2017-03-13T21:43:02Z">\n')
            self.file.write('<osm version="0.6" generator="osm-fieldwork 0.3">\n')
            self.file.flush()

    def footer(self):
        # logging.debug("FIXME: %r" % self.file)
        self.file.write("</osm>\n")
        self.file.flush()
        if self.file is False:
            self.file.close()

    def write(self,
              data = None
              ):
        if type(data) == list:
            if data is not None:
                for line in data:
                    self.file.write("%s\n" % line)
        else:
            self.file.write("%s\n" % data)

    def createWay(self,
                  way: dict,
                  modified: bool = False
                  ):
        """This creates a string that is the OSM representation of a node"""
        attrs = dict()
        osm = ""

        # Add default attributes
        if modified:
            attrs["action"] = "modify"
        if "osm_way_id" in way["attrs"]:
            attrs["id"] = int(way["attrs"]["osm_way_id"])
        elif "osm_id" in way["attrs"]:
            attrs["id"] = int(way["attrs"]["osm_id"])
        elif "id" in way["attrs"]:
            attrs["id"] = int(way["attrs"]["id"])
        else:
            attrs["id"] = self.start
            self.start -= 1
        if "version" not in way["attrs"]:
            attrs["version"] = 1
        else:
            attrs["version"] = way["attrs"]["version"] + 1
        attrs["timestamp"] = datetime.now().strftime("%Y-%m-%dT%TZ")
        # If the resulting file is publicly accessible without authentication, The GDPR applies
        # and the identifying fields should not be included
        if "uid" in way and attrs["uid"] is not None:
            attrs["uid"] = way["attrs"]["uid"]
        if "user" in way and attrs["user"] is not None:
            attrs["user"] = way["attrs"]["user"]

        # Make all the nodes first. The data in the track has 4 fields. The first two
        # are the lat/lon, then the altitude, and finally the GPS accuracy.
        # newrefs = list()
        node = dict()
        node["attrs"] = dict()
        # The geometry is an EWKT string, so there is no need to get fancy with
        # geometries, just manipulate the string, as OSM XML it's only strings
        # anyway.
        # geom = way['geom'][19:][:-2]
        # print(geom)
        # points = geom.split(",")
        # print(points)

        # epdb.st()
        # loop = 0
        # while loop < len(way['refs']):
        #     #print(f"{points[loop]} {way['refs'][loop]}")
        #     node['timestamp'] = attrs['timestamp']
        #     if 'user' in attrs and attrs['user'] is not None:
        #         node['attrs']['user'] = attrs['user']
        #     if 'uid' in attrs and attrs['uid'] is not None:
        #         node['attrs']['uid'] = attrs['uid']
        #     node['version'] = 0
        #     lat,lon = points[loop].split(' ')
        #     node['attrs']['lat'] = lat
        #     node['attrs']['lon'] = lon
        #     node['attrs']['id'] = way['refs'][loop]
        #     osm += self.createNode(node) + '\n'
        #     loop += 1

        # Processs atrributes
        line = ""
        for ref, value in attrs.items():
            line += "%s=%r " % (ref, str(value))
        osm += "  <way " + line + ">"

        for ref in way["refs"]:
            osm += '\n    <nd ref="%s"/>' % ref

        if "tags" in way:
            for key, value in way["tags"].items():
                if value is None:
                    continue
                if key == "track":
                    continue
                if key not in attrs:
                    newkey = self.convert.escape(key)
                    osm += "\n    <tag k='%s' v=%r/>" % (newkey, value)
            if modified:
                osm += (
                    '\n    <tag k="fixme" v="Do not upload this without validation!"/>'
                )
            osm += "\n"

        osm += "  </way>"

        return osm

    def createNode(self,
                   node: dict,
                   modified: bool = False
                   ):
        """This creates a string that is the OSM representation of a node"""
        attrs = dict()
        # Add default attributes
        if modified:
            attrs["action"] = "modify"

        if "id" in node["attrs"]:
            attrs["id"] = int(node["attrs"]["id"])
        else:
            attrs["id"] = self.start
            self.start -= 1
        if "version" not in node["attrs"]:
            attrs["version"] = "1"
        else:
            attrs["version"] = int(node["version"]) + 1
        attrs["lat"] = node["attrs"]["lat"]
        attrs["lon"] = node["attrs"]["lon"]
        attrs["timestamp"] = datetime.now().strftime("%Y-%m-%dT%TZ")
        # If the resulting file is publicly accessible without authentication, THE GDPR applies
        # and the identifying fields should not be included
        if "uid" in node and attrs["uid"] is not None:
            attrs["uid"] = node["uid"]
        if "user" in node and attrs["ser"] is not None:
            attrs["user"] = node["user"]

        # Processs atrributes
        line = ""
        osm = ""
        for ref, value in attrs.items():
            line += "%s=%r " % (ref, str(value))
        osm += "  <node " + line

        if "tags" in node:
            osm += ">"
            for key, value in node["tags"].items():
                if not value:
                    continue
                if key not in attrs:
                    osm += "\n    <tag k='%s' v=%r/>" % (key, value)
            if modified:
                osm += (
                    '\n    <tag k="fixme" v="Do not upload this without validation!"/>'
                )
            osm += "\n  </node>"
        else:
            osm += "/>"

        return osm

    def createTag(self,
                  field: str,
                  value: str
                  ):
        """Create a data structure for an OSM feature tag"""
        newval = str(value)
        newval = newval.replace("&", "and")
        newval = newval.replace('"', "")
        tag = dict()
        # logging.debug("OSM:makeTag(field=%r, value=%r)" % (field, newval))

        newtag = field
        change = newval.split("=")
        if len(change) > 1:
            newtag = change[0]
            newval = change[1]

        tag[newtag] = newval
        return tag

    def loadFile(self,
                 file: str
                 ):
        """Read a OSM XML file generated by osm_fieldwork"""
        size = os.path.getsize(file)
        with open(file, "rb") as file:
            xml = file.read(size)  # Instances are small, read the whole file
            doc = xmltodict.parse(xml)
            if "osm" not in doc:
                logging.warning("No data in this instance")
                return False
            field = doc["osm"]
        for node in field["node"]:
            attrs = {
                "id": node["@id"],
                "lat": node["@lat"],
                "lon": node["@lon"],
                "timestamp": node["@timestamp"],
            }
            tags = dict()
            if "tag" in node:
                for tag in node["tag"]:
                    tags[tag["@k"]] = tag["@v"].strip()
                node = {"attrs": attrs, "tags": tags}
                self.data[node["attrs"]["id"]] = node

    def dump(self):
        """Dump the contents of an OSM file"""
        for id, item in self.data.items():
            for k, v in item["attrs"].items():
                print(f"{k} = {v}")
            for k, v in item["tags"].items():
                print(f"\t{k} = {v}")

    def getFeature(self,
                   id: int
                   ):
        """Get the data for a feature from the loaded OSM data file"""
        return self.data[id]

    def getFields(self):
        """Extract all the tags used in this file"""
        fields = list()
        for id, item in self.data.items():
            keys = list(item["tags"].keys())
            for key in keys:
                if key not in fields:
                    fields.append(key)
        print(fields)


if __name__ == "__main__":
    # Command Line options
    parser = argparse.ArgumentParser(
        description="This program conflates ODK data with existing features from OSM."
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-o", "--osmfile", help="OSM XML file created by Osm-Fieldwork")
    args = parser.parse_args()

    # This program needs options to actually do anything
    if len(argv) == 1:
        parser.print_help()
        quit()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

    osm = OsmFile()
    osm.loadFile(args.osmfile)
    osm.dump()



"""Index of XLSForm file paths."""

import os


xlsforms_path = os.path.dirname(os.path.abspath(__file__))

buildings = f"{xlsforms_path}/buildings.xls"
cemeteries = f"{xlsforms_path}/cemeteries.xls"
education = f"{xlsforms_path}/education.xls"
healthcare = f"{xlsforms_path}/healthcare.xls"
landuse = f"{xlsforms_path}/landuse.xls"
natural = f"{xlsforms_path}/natural.xls"
place = f"{xlsforms_path}/place.xls"
religious = f"{xlsforms_path}/religious.xls"
school = f"{xlsforms_path}/school.xls"
solidwaste = f"{xlsforms_path}/solidwaste.xlsx"
toilets = f"{xlsforms_path}/toilets.xls"
transportation = f"{xlsforms_path}/transportation.xls"
waste_collection = f"{xlsforms_path}/waste_collection.xlsx"
waste = f"{xlsforms_path}/waste.xls"
waterpoints = f"{xlsforms_path}/waterpoints.xls"
waterways = f"{xlsforms_path}/waterways.xls"



#!/usr/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of osm-Fieldwork.
#
#     Osm-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with osm_fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import logging
import sys
import os
import pandas as pd
from geojson import Feature, FeatureCollection
import geojson
from osm_fieldwork.xlsforms import xlsforms_path
import yaml


# Instantiate logger
log = logging.getLogger(__name__)


class FilterData(object):
    def __init__(self,
                 filespec: str = None
                 ):
        self.tags = dict()
        if filespec:
            self.parse(filespec)

    def parse(self,
              filespec: str
              ):
        """Read in the XLSForm and extract the data we want"""
        entries = pd.read_excel(filespec, sheet_name=[0, 1, 2])
        
        title = entries[2]['form_title'].to_list()[0]
        extract = ""
        for entry in entries[0]['type']:
            if str(entry) == 'nan':
                continue
            if entry[:20] == "select_one_from_file":
                extract = entry[21:]
        log.info(f"Got data extract filename: \"{extract}\", title: \"{title}\"")
        total = len(entries[1]['list_name'])
        index = 1
        while index < total:
            key = entries[1]['list_name'][index]
            if key == 'model' or str(key) == "nan":
                index += 1
                continue
            if 'name' in entries:
                value = entries[1]['name'][index]
            else:
                value = None
            if value == "<text>" or str(value) == "null":
                index += 1
                continue
            if key not in self.tags:
                self.tags[key] = list()
            self.tags[key].append(value)
            index += 1

        # The yaml config file for the query has a list of columns
        # to keep in addition to this default set.
        path = xlsforms_path.replace("xlsforms", "data_models")
        category = os.path.basename(filespec).replace(".xls", "")
        file = open(f"{path}/{category}.yaml", "r").read()
        self.yaml = yaml.load(file, Loader=yaml.Loader)
        keep = ("name",
                "name:en",
                "id",
                "operator",
                "addr:street",
                "addr:housenumber",
                "osm_id",
                "title",
                "tags",
                "label",
                "landuse",
                "opening_hours",
                "tourism",
                )
        self.keep = list(keep)
        if 'keep' in self.yaml:
            self.keep.extend(self.yaml['keep'])

        return title, extract

    def cleanData(self,
                  data
                  ):
        tmpfile = data
        if type(data) == str:
            outfile = open(f"new-{data}", "x")
            infile = open(tmpfile, "r")
            indata = geojson.load(infile)
        elif type(data) == bytes:
            indata = eval(data.decode())
        else:
            indata = data
        # these just create noise in the log file
        ignore = (
            "timestamp",
            "version",
            "changeset",
            )
        collection = list()
        for feature in indata['features']:
            properties = dict()
            for key, value in feature['properties'].items():
                if key in self.keep:
                    if key == 'tags':
                        for k, v in value.items():
                            if k[:4] == "name":
                                properties['title'] = value[k]
                                properties['label'] = value[k]
                            else:
                                properties[k] = v
                    else:
                        if key == 'osm_id':
                            properties['id'] = value
                        else:
                            properties[key] = value
                else:
                    if key in self.tags.keys():
                        if key == "name":
                            properties['title'] = self.tags[key]
                            properties['label'] = self.tags[key]
                        if value in self.tags[key]:
                            properties[key] = value
                        else:
                            if value != "yes":
                                log.warning(f"Value {value} not in the data model!")
                            continue
                    else:
                        if key in ignore:
                            continue
                        log.warning(f"Tag {key} not in the data model!")
                        continue
            newfeature = Feature(geometry=feature['geometry'], properties=properties)
            collection.append(newfeature)
        if type(data) == str:
            geojson.dump(FeatureCollection(collection), outfile)
        return FeatureCollection(collection)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert ODK XML instance file to CSV format"
    )
    parser.add_argument("-v", "--verbose", nargs="?", const="0", help="verbose output")
    parser.add_argument("-i", "--infile", help="The data extract for ODK Collect")
    parser.add_argument("-x", "--xform", help="The XForm for ODK Collect")
    parser.add_argument("-o", "--outfile", default="models.yaml", help="The Yaml file of all tags and values")
    args = parser.parse_args()

    # if verbose, dump to the termina
    if not args.verbose:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        root.addHandler(ch)

    xls = FilterData()
    path = xlsforms_path.replace("xlsforms", "data_models")
    models = FilterData()
    if not args.xform:
        data = models.parse(f"{path}/Impact Areas - Data Models V1.1.xlsx")
    else:
        data = models.parse(args.xform)
    if args.infile:
        cleaned = models.cleanData(args.infile)
    else:
        if os.path.exists(args.outfile):
            os.remove(args.outfile)
        outfile = open(args.outfile, "w")
        current = None
        tab = "    "
        outfile.write("""
        # Do not edit this file, it is generated by the filter_data.py
        # script from the XForm spreadsheet.
        """)
        for key, value in data.items():
            if key != current:
                outfile.write(f"\n{key}:\n")
                for val in value:
                    outfile.write(f"{tab}- {val}\n")




#!/usr/bin/python3

# Copyright (c) 2022 Humanitarian OpenStreetMap Team
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import logging
from sys import argv
from osgeo import ogr
from osm_fieldwork.osmfile import OsmFile

# Instantiate logger
log = logging.getLogger(__name__)


class InputFile(object):
    def __init__(self, source=None):
        """Initialize Input Layer"""
        self.datain = None
        self.source = source
        if source[0:3] == "pg:":
            logging.info("Opening database connection to: %s" % source)
            connect = "PG: dbname=" + source[3:]
            # if options.get('dbhost') != "localhost":
            # connect += " host=" + options.get('dbhost')
            self.datain = ogr.Open(connect)
        else:
            logging.info("Opening data file: %s" % source)
            self.datain = ogr.Open(source)

        # Copy the data into memory for better performance
        memdrv = ogr.GetDriverByName("MEMORY")
        self.msmem = memdrv.CreateDataSource("msmem")
        self.msmem.CopyLayer(self.datain.GetLayer(), "msmem")
        self.layer = self.msmem.GetLayer()
        self.fields = self.datain.GetLayer().GetLayerDefn()
        self.memlayer = None
        self.mem = None
        self.tags = dict()
        for i in range(self.fields.GetFieldCount()):
            self.tags[self.fields.GetFieldDefn(i).GetName()] = None

    def clip(self, boundary=None):
        """Clip a data source by a boundary"""
        if boundary:
            logging.info("Clipping %s using %s" % (self.source, boundary))
            memdrv = ogr.GetDriverByName("MEMORY")
            self.mem = memdrv.CreateDataSource("data")
            self.memlayer = self.mem.CreateLayer("data", geom_type=ogr.wkbUnknown)

            poly = ogr.Open(boundary)
            layer = poly.GetLayer()
            ogr.Layer.Clip(self.layer, layer, self.memlayer)

    def mergeTags(self, osm=None, odk=None):
        """Merge two sets of tags together"""
        if str(osm) < str(odk):
            # logging.debug("No changes made to tags")
            return None
        else:
            # logging.debug("Changes tags")
            osm.update(odk)
            return osm
        return None

    def getFeature(self, data=None):
        id = data["attrs"]["id"]
        result = None
        if not id or int(id) < 0:
            if "name" in data["tags"]:
                query = f"SELECT *, ST_AsEWKT(geom) AS wkt FROM nodes WHERE tags->>'name'='{data['tags']['name']}'"
                # print(query)
                result = self.datain.ExecuteSQL(query)
                if result.GetFeatureCount() == 0:
                    logging.debug(
                        f"Feature not found in nodes for ID: {data['tags']['name']}"
                    )
                    result = None
                else:
                    logging.debug(f"Feature {data['tags']['name']} found in nodes")
            else:
                return None

        if not result:
            query = (
                f"SELECT *, ST_AsEWKT(geom) AS wkt FROM ways_poly WHERE osm_id='{id}'"
            )
            # print(query)
            result = self.datain.ExecuteSQL(query)
            if result.GetFeatureCount() == 0:
                # print(data)
                if "name" in data["tags"]:
                    query = f"SELECT *, ST_AsEWKT(geom) AS wkt FROM nodes WHERE tags->>'name'='{data['tags']['name']}'"
                    # print(query)
                    result = self.datain.ExecuteSQL(query)
                    if result.GetFeatureCount() == 0:
                        logging.debug(
                            f"Feature not found in ways for ID: {data['tags']['name']}"
                        )
                    else:
                        logging.debug(
                            f"Feature found in ways for ID: {data['tags']['name']}"
                        )
                return None

        # There is only one feature for an OSM ID
        feature = result.GetFeature(0)
        if feature.GetFieldCount() == 0:
            logging.debug(f"No fields found in feature for ID: {id}")
            return None

        index = feature.GetFieldIndex("tags")
        tags = eval(feature.GetField(index))

        index = feature.GetFieldIndex("version")
        version = feature.GetField(index)

        # Refs are stored as a string with a colon delimiter
        index = feature.GetFieldIndex("refs")
        if index < 0:
            refs = list()
        else:
            refs = eval(feature.GetField(index))

        # There should only be one feature returned from the query
        index = feature.GetFieldIndex("wkt")
        geom = feature.GetField(index)
        # geom = result[0].GetGeometryRef()

        # a ways don't use lat-lon, it uses references to nodes instead
        attrs = {"id": id, "version": version}
        self.tags = dict()
        for k, v in tags.items():
            self.tags[k] = v
        feature = {"tags": self.tags, "attrs": attrs, "refs": refs, "geom": geom}
        return feature

    def dump(self):
        """Dump internal data"""
        print(f"Data source is: {self.source}")
        if self.layer:
            print("There are %d in the layer" % self.layer.GetFeatureCount())
        if self.memlayer:
            print("There are %d in the layer" % self.memlayer.GetFeatureCount())


if __name__ == "__main__":
    # Command Line options
    parser = argparse.ArgumentParser(
        description="This program conflates ODK data with existing features from OSM."
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument(
        "-c", "--odkfile", help="ODK CSV file downloaded from ODK Central"
    )
    parser.add_argument("-f", "--osmfile", help="OSM XML file created by Osm-Fieldwork")
    parser.add_argument(
        "-o", "--outfile", default="tmp.osm", help="Output file from the merge"
    )
    parser.add_argument(
        "-b", "--boundary", help="Boundary polygon to limit the data size"
    )
    args = parser.parse_args()

    # This program needs options to actually do anything
    if len(argv) == 1:
        parser.print_help()
        quit()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

    # This is the existing OSM data, a database or a file
    if args.osmfile:
        osmf = InputFile(args.osmfile)
        # osmf.dump()
        if args.boundary:
            osmf.clip(args.boundary)
        osmf.dump()
    else:
        logging.error("No OSM data source specified!")
        parser.print_help()
        quit()

    # Create an OSM XML handler, which writes to the output file
    odkf = OsmFile(filespec=args.outfile)
    # And also loads the POIs from the ODK Central submission
    odkf.loadFile(args.odkfile)

    for id in odkf.data:
        # print(odkf.data[id])
        feature = osmf.getFeature(odkf.data[id])
        out = list()
        if not feature:
            # logging.debug(f"No feature found for ID {id}")
            # feature = osmf.createFeature(odkf.data[id])
            out.append(odkf.createNode(odkf.data[id], modified=True))
        else:
            tags = osmf.mergeTags(feature["tags"], odkf.data[id]["tags"])
            if tags:
                odkf.data[id]["tags"] = tags
            if "name" in feature["tags"]:
                # if tags and odkf.data[id]['tags']['name'] != feature['tags']['name']:
                if tags:
                    out.append(odkf.createNode(odkf.data[id], modified=True))
            else:
                out.append(odkf.createWay(feature, modified=True))
        odkf.write(out)

    # # FIXME: for now just copy the data file from Central
    # # to test input parsing, and output accuracy.
    # out = list()
    # for id, node in odkf.data.items():
    #     out.append(odkf.createNode(node, modified=True))
    # odkf.write(out)
    odkf.footer()
    logging.info("Wrote %s: " % args.outfile)

# osmoutfile = os.path.basename(args.infile.replace(".csv", ".osm"))
# csvin.createOSM(osmoutfile)



#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of Osm-Fieldwork.
#
#     Osm-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Osm-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import yaml
import argparse
import logging
import sys

"""This parses a yaml file into a dictionary for easy access."""

# Instantiate logger
log = logging.getLogger(__name__)


class YamlFile(object):
    """Config file in YAML format"""

    def __init__(self, data):
        self.filespec = None
        #if data == str:
        self.filespec = data
        self.file = open(data, "rb").read()
        self.yaml = yaml.load(self.file, Loader=yaml.Loader)
        #else:
        #    self.yaml = yaml.load(str(data), Loader=yaml.Loader)

    def privateData(self, keyword: str):
        """See if a keyword is in the private data category"""
        return keyword in self.yaml["private"]

    def ignoreData(self, keyword: str):
        """See if a keyword is in the ignore data category"""
        return keyword in self.yaml["ignore"]

    def tagsData(self, keyword: str):
        """See if a keyword is in the tags completness section"""
        return keyword in self.yaml["tags"]

    def hasList(self, keyword: str):
        for tags in self.yaml["tags"]:
            for tag in tags:
                if tag == keyword:
                    return tag
            logging.debug(type(tags))

    def dump(self):
        """Dump the contents of the yaml file"""
        if self.filespec:
            print("YAML file: %s" % self.filespec)
        for item in self.yaml:
            print(f"{item}")

    def write(self, table: list, where: list):
        tab = "    "
        yaml = ["select:", f"{tab}\"osm_id\": id", f"{tab}tags:"]
        for item in where:
            yaml.append(f"{tab}{tab}- {item}")
        yaml.append("from:")
        for item in table:
            yaml.append(f"{tab}- {item}")
        yaml.append("where:")
        yaml.append(f"{tab}tags:")
        notnull = f"{tab}{tab}- " + "{"
        for item in where:
            notnull += f"{item}: NOT NULL, "
        notnull = f"{notnull[:-2]}"
        notnull += "}"
        yaml.append(f"{notnull}")
        return yaml

#
# This script can be run standalone for debugging purposes. It's easier to debug
# this way than using pytest,
#
if __name__ == "__main__":
    # Enable logging to the terminal by default
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    root.addHandler(ch)

    parser = argparse.ArgumentParser(description="Read and parse a YAML file")
    parser.add_argument(
        "-i", "--infile", default="./xforms.yaml", help="The YAML input file"
    )
    args = parser.parse_args()

    yaml1 = YamlFile(args.infile)

    table = ("nodes", "ways_poly")
    where = ("building", "amenity", "shop", "tourism")
    tmp = yaml1.write(table, where)
    yaml2 = YamlFile(tmp)
    yaml2.dump()



#!/usr/bin/python3

# Copyright (c) 2021, 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of Osm-Fieldwork.
#
#     Osm-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Osm-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import os
import sys
import argparse
from osm_fieldwork.convert import Convert

parser = argparse.ArgumentParser(
    description="Read and convert a CSV file from ODK Central"
)
parser.add_argument("--infile", default="tests/test.csv", help="The CSV input file")
args = parser.parse_args()

csv = Convert("xforms.yaml")

def test_get_keyword():
    """Convert a feature"""
    if "sac_scale" in csv.yaml.yaml["tags"]:
        assert 0
    else:
        assert 1


def test_no_keyword():
    """Convert a feature that doesn't exist"""
    if "sac_scale" in csv.yaml.yaml["convert"]:
        csv.getKeyword("doesn't exist")
        assert 0
    else:
        assert 1


# def test_convert_list():
#     """Convert a list of features"""
#     features = list()
#     features.append("firepit")
#     features.append("parking")
#     features.append("viewpoint")
#     result = csv.convertList(features)
#     assert result[0]['leisure'] == "firepit" and result[1]['amenity'] == "parking" and result[2]['tourism'] == "viewpoint"

# def test_bool_keyword():
#     """Convert a feature"""
#     result = csv.getKeyword("bear box")
#     assert result == "bear box"


def test_convert_tag():
    tmp1 = csv.convertTag("altitude")
    tmp2 = csv.convertTag("foobar")
    assert tmp1 == "ele" and tmp2 == "foobar"


# Run standalone for easier debugging when not under pytest
if __name__ == "__main__":
    test_get_keyword()
    test_no_keyword()
    #    test_convert_list()
    #    test_bool_keyword()
    test_convert_tag()



#!/usr/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of osm_fieldwork.
#
#     Underpass is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with osm_fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import os
import sys

sys.path.append(f"{os.getcwd()}/osm_fieldwork")
import argparse
from osmfile import OsmFile

parser = argparse.ArgumentParser(
    description="Read and parse a CSV file from ODK Central"
)
parser.add_argument("--infile", default="tests/test.csv", help="The CSV input file")
args = parser.parse_args()

# Delete the test output file
if os.path.exists("tests/test.osm"):
    os.remove("tests/test.osm")

osm = OsmFile("tests/test.osm")


def test_init():
    """Make sure the OSM file is initialized"""
    assert os.path.exists("tests/test.osm")


def test_header():
    osm.header()
    assert os.stat("tests/test.osm").st_size > 0


def test_footer():
    osm.footer()
    tmp = open("tests/test.osm", "r")
    lines = tmp.readlines()
    for line in lines:
        last = line
    assert last == "</osm>\n"


def test_create_tag():
    tmp = osm.createTag("foo", "bar")
    assert tmp["foo"] == "bar"


def test_create_node_notags():
    node = dict()
    attrs = dict(id=12345, lat=1, lon=2, uid=54321, user="bar")
    node["attrs"] = attrs
    tmp = osm.createNode(node).lstrip().split(" ")
    assert tmp[1] == "id='12345'" and tmp[2] == "version='1'" and tmp[3] == "lat='1'"
    # print(tmp)


def test_create_node_modified():
    node = dict()
    attrs = dict(id=12345, lat=1, lon=2, version=2, uid=54321, user="bar")
    node["attrs"] = attrs
    node["version"] = 7
    tmp = osm.createNode(node, modified=True)
    assert tmp.find("action")
    # print(tmp)


def test_create_node_tags():
    node = dict()
    attrs = dict(id=12345, lat=1, lon=2, uid=54321, user="bar")
    node["attrs"] = attrs
    tmp = osm.createNode(node)
    assert tmp.find("bar") and tmp.find("foo")
    # print(tmp)


def test_create_way():
    way = dict()
    way = dict()
    attrs = dict(id=12345, lat=1, lon=2, uid=54321, user="bar")
    way["attrs"] = attrs
    tags = dict(foo="bar", bar="foo")
    refs = list()
    refs.append(12345)
    refs.append(23456)
    refs.append(34567)
    way["tags"] = tags
    way["refs"] = refs
    osm.createWay(way)
    # print(tmp)


if __name__ == "__main__":
    test_init()
    test_header()
    test_footer()
    test_create_tag()
    test_create_node_notags()
    test_create_node_tags()
    test_create_way()
    test_create_node_modified()
    # Delete the test output file
    if os.path.exists("tests/test.osm"):
        os.remove("tests/test.osm")



#!/usr/bin/python3

# Copyright (c) 2022 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import os
import logging
import sys
import re
import yaml
import json
from geojson import Point, Feature, FeatureCollection, dump, Polygon
import geojson
from osm_fieldwork.filter_data import FilterData
from osm_fieldwork.xlsforms import xlsforms_path
import requests
# from requests.auth import HTTPBasicAuth
from io import BytesIO
import zipfile
import time
import psycopg2
from shapely.geometry import shape, Polygon
import overpy
import shapely
from shapely import wkt


def getChoices():
    """Get the categories and associated XLSFiles fgrom the config file"""
    data = dict()
    path = xlsforms_path.replace("xlsforms", "data_models")
    if os.path.exists(f"{path}/category.yaml"):
        file = open(f"{path}/category.yaml", "r").read()
        contents = yaml.load(file, Loader=yaml.Loader)
        for entry in contents:
            [[k,v]] = entry.items()
            data[k] = v[0]
    return data

# Instantiate logger
log = logging.getLogger(__name__)


class DatabaseAccess(object):
    def __init__(self,
                 dbhost: str = None,
                 dbname: str = None
                 ):
        self.dbshell = None
        self.dbcursor = None
        self.category = None
        if dbname == "underpass":
            # Authentication data
            # self.auth = HTTPBasicAuth(self.user, self.passwd)

            # Use a persistant connect, better for multiple requests
            self.session = requests.Session()
            self.url = "https://raw-data-api0.hotosm.org/v1"
            self.headers = {"accept": "application/json", "Content-Type": "application/json"}
        else:
            logging.info("Opening database connection to: %s" % dbhost)
            connect = "PG: dbname=" + dbname
            if dbhost is None or dbhost == "localhost":
                connect = f"dbname={dbname}"
            else:
                connect = f"host={dbhost} dbname={dbname}"
            try:
                self.dbshell = psycopg2.connect(connect)
                self.dbshell.autocommit = True
                self.dbcursor = self.dbshell.cursor()
                if self.dbcursor.closed == 0:
                    logging.info(f"Opened cursor in {dbname}")
            except Exception as e:
                logging.error("Couldn't connect to database: %r" % e)

    def createJson(self,
                   category: str,
                   boundary,
                   poly: bool = False
                   ):
        path = xlsforms_path.replace("xlsforms", "data_models")
        file = open(f"{path}/{category}.yaml", "r").read()
        data = yaml.load(file, Loader=yaml.Loader)

        features = dict()
        features['geometry'] = boundary

        # The database tables to query
        # if tags exists, then only query those fields
        columns = dict()
        tags = data['where']['tags'][0]
        for tag, value in tags.items():
            if value == "not null":
                columns[tag] = []
            else:
                columns[tag] = value
        filters = {"tags": {"all_geometry": {"join_or": columns}}}
        features['filters'] = filters
        tables = list()
        for table in data['from']:
            if table == "nodes":
                tables.append("point")
            elif table == "ways_poly":
                tables.append("polygon")
            elif table == "ways_line":
                tables.append("line")
            elif table == "relations":
                tables.append("line")
        features["geometryType"] = tables
        if not poly:
            features["centroid"] = "true"
        return json.dumps(features)

    def createSQL(self,
                  category: str,
                  polygon: bool = False
                  ):
        path = xlsforms_path.replace("xlsforms", "data_models")
        file = open(f"{path}/{category}.yaml", "r").read()
        data = yaml.load(file, Loader=yaml.Loader)

        sql = list()
        # The database tables to query
        tables = data['from']
        for table in tables:
            query = "SELECT "
            select = data['select']
            if polygon:
                centroid = "geom"
            else:
                centroid = "ST_Centroid(geom)"
            # if tags exists, then only return those fields
            if 'tags' in select:
                for tag in select['tags']:
                    query += f" {select[tag]} AS {tag}, "
                query += f"osm_id AS id, ST_AsEWKT({centroid} "
            else:
                query += f"osm_id AS id, ST_AsEWKT({centroid}), tags "
            query += f" FROM {table} "
            where = data['where']
            # if tags exists, then only query those fields
            if 'where' in data:
                query += " WHERE "
                tags = data['where']['tags'][0]
                for tag, value in tags.items():
                    if value == "not null":
                        query += f"tags->>\'{tag}\' IS NOT NULL OR "
                    else:
                        # in the yaml file, multiple values for the same tag
                        # are a list
                        for val in value:
                            query += f"tags->>\'{tag}\'=\'{val}\' OR "
            sql.append(query[:-4])
        return sql

    def queryLocal(self,
                   query: str = None,
                   ewkt: str = None
                   ):
        sql = f"DROP VIEW IF EXISTS ways_view;CREATE TEMP VIEW ways_view AS SELECT * FROM ways_poly WHERE ST_CONTAINS(ST_GeomFromEWKT('SRID=4326;{ewkt.wkt}'), geom)"
        self.dbcursor.execute(sql)
        sql = f"DROP VIEW IF EXISTS nodes_view;CREATE TEMP VIEW nodes_view AS SELECT * FROM nodes WHERE ST_CONTAINS(ST_GeomFromEWKT('SRID=4326;{ewkt.wkt}'), geom)"
        self.dbcursor.execute(sql)

        sql = f"DROP VIEW IF EXISTS relations_view;CREATE TEMP VIEW relations_view AS SELECT * FROM nodes WHERE ST_CONTAINS(ST_GeomFromEWKT('SRID=4326;{ewkt.wkt}'), geom)"
        # self.dbcursor.execute(sql)

        if query.find(" ways_poly ") > 0:
            query = query.replace("ways_poly", "ways_view")
        elif query.find(" nodes ") > 0:
            query = query.replace("nodes", "nodes_view")
        features = list()
        log.debug(query)
        self.dbcursor.execute(query)
        result = self.dbcursor.fetchall()
        logging.info("Query returned %d records" % len(result))
        for item in result:
            gps = item[1][16:-1].split(" ")
            if len(gps) == 2:
                poi = Point((float(gps[0]), float(gps[1])))
            else:
                gps = item[1][10:]
                poi = wkt.loads(gps)
            tags = item[2]
            tags["id"] = item[0]
            if "name:en" in tags:
                tags["title"] = tags["name:en"]
                tags["label"] = tags["name:en"]
            elif "name" in tags:
                tags["title"] = tags["name"]
                tags["label"] = tags["name"]
            else:
                tags["title"] = tags["id"]
                tags["label"] = tags["id"]
            features.append(Feature(geometry=poi, properties=tags))
        return features

    def queryRemote(self,
                    query: str = None
                    ):
        url = f"{self.url}/snapshot/"
        result = self.session.post(url, data=query, headers=self.headers)
        if result.status_code != 200:
            log.error(f"{result.json()['detail'][0]['msg']}")
            return None
        task_id = result.json()['task_id']
        newurl = f"{self.url}/tasks/status/{task_id}"
        while True:
            result = self.session.get(newurl, headers=self.headers)
            if result.json()['status'] == "PENDING":
                log.debug("Retrying...")
                time.sleep(1)
            elif result.json()['status'] == "SUCCESS":
                break
        zip = result.json()['result']['download_url']
        result = self.session.get(zip, headers=self.headers)
        fp = BytesIO(result.content)
        zfp = zipfile.ZipFile(fp, "r")
        zfp.extract("Export.geojson", "/tmp/")
        # Now take that taskid and hit /tasks/status url with get
        data = zfp.read("Export.geojson")
        os.remove("/tmp/Export.geojson")
        return eval(data)
    #   return zfp.read("Export.geojson")

class PostgresClient(DatabaseAccess):
    """Class to handle SQL queries for the categories"""
    def __init__(self,
                 dbhost: str = None,
                 dbname: str = None,
                 output: str = None
    ):
        """Initialize the postgres handler"""
        # OutputFile.__init__( self, output)
        self.boundary = None
        super().__init__(dbhost, dbname)

    def getFeatures(self,
                    boundary,
                    filespec: str,
                    polygon: bool,
                    category: str,
                    xlsfile: str,
                    ):
        """Extract buildings from Postgres"""
        logging.info("Extracting features from Postgres...")

        if type(boundary) != dict:
            clip = open(boundary, "r")
            geom = geojson.load(clip)
            if 'features' in geom:
                poly = geom['features'][0]['geometry']
            else:
                poly = geom["geometry"]
        else:
            poly = boundary
        wkt = shape(poly)

        if len(xlsfile) > 0:
            config = xlsfile.replace(".xls", "")
        else:
            config = category
        if self.dbshell:
            # features = list()
            sql = self.createSQL(config, polygon)
            all = list()
            for query in sql:
                result = self.queryLocal(query, wkt)
                all.extend(result)
            collection = FeatureCollection(all)
        else:
            request = self.createJson(config, poly, polygon)
            collection = self.queryRemote(request)
        if not collection:
            return None

        if len(collection['features']) == 0:
            tags = { 'title': category, 'label': category, 'id': 0}
            center = shapely.centroid(wkt)
            feature = [Feature(geometry=center, properties=tags)]
            new = FeatureCollection(feature)
            extract = "yes"
        else:
            # Process the XLSForm source file and scan it for valid tags
            # and values.
            cleaned = FilterData()
            models = xlsforms_path.replace("xlsforms", "data_models")
            if not xlsfile:
                xlsfile = f"{category}.xls"
            file = f"{xlsforms_path}/{xlsfile}"
            if os.path.exists(file):
                title, extract = cleaned.parse(file)
            elif os.path.exists(f"{file}x"):
                title, extract = cleaned.parse(f"{file}x")
            # Remove anything in the data extract not in the choices sheet.
            new = cleaned.cleanData(collection)

        # This will be set if the XLSForm contains a select_one_from_file
        if len(extract) > 0:
            # filespec = f"/tmp/{outfile}"
            jsonfile = open(filespec, "w")
            dump(new, jsonfile)
            jsonfile.close()
        return new

class OverpassClient(object):
    """Class to handle Overpass queries"""
    def __init__(self,
                 output: str = None):
        """Initialize Overpass handler"""
        self.overpass = overpy.Overpass()
        #OutputFile.__init__(self, output)

    def getFeatures(self,
                    boundary,
                    filespec: str,
                    xlsfile: str,
                    category: str
                    ):
        """Extract buildings from Overpass"""
        logging.info("Extracting features...")

        poly = ""
        if type(boundary) == str:
            clip = open(boundary, "r")
            geom = geojson.load(clip)
            if 'features' in geom:
                aoi = geom['features'][0]['geometry']
            else:
                aoi = geom["geometry"]
            wkt = shape(aoi)
            lat, lon = wkt.exterior.coords.xy
            index = 0
            while index < len(lat):
                poly += f"{lon[index]} {lat[index]} "
                index += 1
        else:
            for coords in boundary['geometry']['coordinates'][0]:
                poly += f"{coords[1]} {coords[0]}"

        query = (f'[out:json];way[\"building\"](poly:\"{poly[:-1]}\");(._;>;);out body geom;')
        result = self.overpass.query(query)
        features = list()
        for way in result.ways:
            poly = list()
            for coords in way.attributes['geometry']:
                lat = coords['lat']
                lon = coords['lon']
                point = [lon, lat]
                poly.append(point)
            exterior = Polygon(poly)
            center = shapely.centroid(exterior)
            features.append(Feature(geometry=center, properties=way.tags))

        collection = FeatureCollection(features)

        cleaned = FilterData()
        file = f"{xlsforms_path}/{xlsfile}"
        if os.path.exists(file):
            cleaned.parse(file)
        else:
            cleaned.parse(f"{file}x")
        new = cleaned.cleanData(collection)
        jsonfile = open(filespec, "w")
        dump(new, jsonfile)
        return new

class FileClient(object):
    """Class to handle Overpass queries"""

    def __init__(self,
                 infile: str,
                 output: str
                 ):
        """Initialize Overpass handler"""
        OutputFile.__init__(self, output)
        self.infile = infile

    def getFeatures(self,
                    boundary,
                    infile: str,
                    outfile: str
                    ):
        """Extract buildings from a disk file"""
        logging.info("Extracting buildings from %s..." % infile)
        if boundary:
            poly = ogr.Open(boundary)
            layer = poly.GetLayer()
            ogr.Layer.Clip(osm, layer, memlayer)
        else:
            layer = poly.GetLayer()

        tmp = ogr.Open(infile)
        layer = tmp.GetLayer()

        layer.SetAttributeFilter("tags->>'building' IS NOT NULL")


if __name__ == "__main__":
    choices = getChoices()
    
    parser = argparse.ArgumentParser(
        description="Make GeoJson data file for ODK from OSM"
    )
    parser.add_argument("-v", "--verbose", nargs="?", const="0", help="verbose output")
    parser.add_argument(
        "-o", "--overpass", action="store_true", help="Use Overpass Turbo"
    )
    parser.add_argument(
        "-p", "--postgres", action="store_true", help="Use a postgres database"
    )
    parser.add_argument(
        "-po", "--polygon", action="store_true", default=False,  help="Output polygons instead of centroids"
    )
    parser.add_argument(
        "-g", "--geojson", help="Name of the GeoJson output file"
    )
    parser.add_argument("-i", "--infile", help="Input data file")
    parser.add_argument("-dn", "--dbname", help="Database name")
    parser.add_argument("-dh", "--dbhost", default="localhost", help="Database host")
    parser.add_argument(
        "-b", "--boundary", help="Boundary polygon to limit the data size"
    )
    parser.add_argument(
        "-c",
        "--category",
        default="buildings",
        choices=choices,
        help="Which category to extract",
    )
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        root = logging.getLogger()
        log.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        root.addHandler(ch)

    if args.geojson == "tmp.geojson":
        # The data file name is in the XML file
        regex = r"jr://file.*\.geojson"
        outfile = None
        filename = args.category + ".xml"
        if not os.path.exists(filename):
            logging.error("Please run xls2xform or make to produce %s" % filename)
            quit()
        with open(filename, "r") as f:
            txt = f.read()
            match = re.search(regex, txt)
            if match:
                tmp = match.group().split("/")
        outfile = tmp[3]
    else:
        outfile = args.geojson.lower()

    xlsfile = choices[args.category]
    if args.postgres:
        logging.info("Using a Postgres database for the data source")
        pg = PostgresClient(args.dbhost, args.dbname, outfile)
        if args.geojson:
            extract = args.geojson
        else:
            infile = FilterData(xlsfile)
            extract = infile.metadata[1]
        pg.getFeatures(args.boundary, extract, args.polygon, args.category, xlsfile)
        log.info(f"Created /tmp/{outfile} for {args.category}")
        # pg.cleanup(outfile)
    elif args.overpass:
        logging.info("Using Overpass Turbo for the data source")
        op = OverpassClient(outfile)
        clip = open(args.boundary, "r")
        geom = geojson.load(clip)
        #op.getFeatures(args.boundary, args.geojson, args.category)
        op.getFeatures(geom, args.geojson, xlsfile, args.category)
    elif args.infile:
        f = FileClient(args.infile)
        f.getFeatures(args.boundary, args.geojson, args.category)
        logging.info("Using file %s for the data source" % args.infile)
    else:
        logging.error("You need to supply either --overpass or --postgres")

        logging.info("Wrote output data file to: %s" % outfile)



#!/usr/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of osm_fieldwork.
#
#     Underpass is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with osm_fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import os
import sys
from  osm_fieldwork.CSVDump import CSVDump

parser = argparse.ArgumentParser(
    description="Read and parse a CSV file from ODK Central"
)
parser.add_argument("--infile", default="tests/test.csv", help="The CSV input file")
args = parser.parse_args()
csv = CSVDump()
data = csv.parse("tests/test.csv")
# print(data)


def test_csv():
    """Make sure the CSV file got loaded and parsed"""
    assert len(data) > 0


def test_init():
    """Make sure the YAML file got loaded"""
    assert len(csv.yaml.yaml) > 0


def test_osm_entry():
    csv.createOSM("tests/test.osm")
    line = {
        "timestamp": "2021-09-25T14:27:43.862Z",
        "end": "2021-09-24T17:55:26.194-06:00",
        "today": "2021-09-24",
        "features": "firepit parking caravans",
        "internet": "none",
        "lat": "38.3697403",
        "lon": "-106.3078813",
        "ele": "2825.998",
        "uid": "123435",
        "user": "Foobar",
    }
    csv.createEntry(line)
    # assert tmp


if __name__ == "__main__":
    test_init()
    test_csv()
    test_osm_entry()



#!/usr/bin/python3

# Copyright (c) 2023 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import csv
import os
import logging
import sys
import json
import geojson
from sys import argv
from osm_fieldwork.convert import Convert
from osm_fieldwork.osmfile import OsmFile
from osm_fieldwork.xlsforms import xlsforms_path
from geojson import Point, Feature, FeatureCollection, dump
from pathlib import Path
#import pandas as pd
import math
import re

# set log level for urlib
log = logging.getLogger(__name__)

class JsonDump(Convert):
    """A class to parse the JSON files from ODK Central or odk2geojson"""

    def __init__(self,
                 yaml: str = None
                 ):
        self.fields = dict()
        self.nodesets = dict()
        self.data = list()
        self.osm = None
        self.json = None
        self.features = list()
        path = xlsforms_path.replace("xlsforms", "")
        if yaml:
            file = f"{path}{yaml}"
        else:
            file = f"{path}/xforms.yaml"
        self.config = super().__init__(yaml)

        self.ignore = ('@id',
                       '@xmlns:ev',
                       '@xmlns:orx',
                       '@xmlns:odk',
                       '@xmlns:h',
                       '@xmlns:jr',
                       '@xmlns:xsd',
                       'start',
                       'end',
                       'today',
                       'meta',
                       )

    # def parseXLS(self, xlsfile: str):
    #     """Parse the source XLSFile if available to look for details we need"""
    #     if xlsfile is not None and len(xlsfile) > 0:
    #         entries = pd.read_excel(xlsfile, sheet_name=[0])
    #         # There will only be a single sheet
    #         names = entries[0]['name']
    #         defaults = entries[0]['default']
    #         total = len(names)
    #         i = 0
    #         while i < total:
    #             entry = defaults[i]
    #             if str(entry) != 'nan':
    #                 pat = re.compile("..last-saved.*")
    #                 if pat.match(entry):
    #                     name = entry.split('#')[1][:-1]
    #                     self.saved[name] = None
    #                 else:
    #                     self.defaults[names[i]] = entry
    #             i += 1
    #     return True

    def createOSM(self,
                  filespec: str = "tmp.osm"
                  ):
        """Create an OSM XML output files"""
        log.debug("Creating OSM XML file: %s" % filespec)
        self.osm = OsmFile(filespec)

    def writeOSM(self,
                 feature: dict
                 ):
        """Write a feature to an OSM XML output file"""
        out = ""
        if "id" in feature["tags"]:
            feature["id"] = feature["tags"]["id"]
        if "lat" not in feature["attrs"] or "lon" not in feature["attrs"]:
            return None
        if "refs" not in feature:
            out += self.osm.createNode(feature)
        else:
            out += self.osm.createWay(feature)
        self.osm.write(out)

    def finishOSM(self):
        """Write the OSM XML file footer and close it"""
        self.osm.footer()

    def createGeoJson(self, file="tmp.geojson"):
        """Create a GeoJson output file"""
        log.debug("Creating GeoJson file: %s" % file)
        self.json = open(file, "w")

    def writeGeoJson(self,
                     feature: dict
                     ):
        """Write a feature to a GeoJson output file"""
        # These get written later when finishing , since we have to create a FeatureCollection
        if "lat" not in feature["attrs"] or "lon" not in feature["attrs"]:
            return None
        self.features.append(feature)

    def finishGeoJson(self):
        """Write the GeoJson FeatureCollection to the output file and close it"""
        features = list()
        for item in self.features:
            #poi = Point()
            poi = Point((float(item["attrs"]["lon"]), float(item["attrs"]["lat"])))
            if "private" in item:
                props = {**item["tags"], **item["private"]}
            else:
                props = item["tags"]
            features.append(Feature(geometry=poi, properties=props))
        collection = FeatureCollection(features)
        dump(collection, self.json)

    def getAllTags(self,
                   data
                   ):
        all_tags = dict()
        tags = dict()
        if type(data) == dict:
            for k, v in data.items():
                if type(v) == dict:
                    # log.debug(f"Processing tag {k} = {v}")
                    for k1, v1 in v.items():
                        # log.debug(f"Processing sub tag {k} = {v}")
                        # tags.update(self.getAllTags(v))
                        if type(v1) == dict:
                            for i, j in v1.items():
                                if type(j) == dict:
                                    # FIXME: this should handle more than one
                                    # but so far I've only it to be accuracy, no other
                                    # tags
                                    k2 = list(j.keys())[0]
                                    tags[k2] = list(j.values())[0]
                                else:
                                    tags[i] = j
                        else:
                            tags[k1] = v1
                        # log.debug(f"TAGS: {tags}")
                else:
                    if v:
                        # if k in self.saved:
                        #     if str(v) == 'nan' or len(v) == 0:
                        #         log.debug(f"FIXME: {k} {v}")
                        #         val = self.saved[k]
                        #         if val and len(v) == 0:
                        #             log.warning(f"Using last saved value for \"{k}\"! Now \"{val}\"" )
                        #             value = val
                        #         else:
                        #             self.saved[k] = value
                        #             log.debug(f"Updating last saved value for \"{k}\" with \"{value}\"")
                        if type(v) != str:
                            tags[k] = str(v)
                        else:
                            tags[k] = v
                all_tags.update(tags)
        return all_tags

    def parse(self,
              filespec: str,
              data: str = None
              ):
        """Parse the JSON file from ODK Central and convert it to a data structure"""
        all_tags = list()
        if not data:
            file = open(filespec, "r")
            infile = Path(filespec)
            if infile.suffix == ".geojson":
                reader = geojson.load(file)
            elif infile.suffix == ".json":
                reader = json.load(file)
            else:
                log.error("Need to specify a JSON or GeoJson file!")
                return all_tags
        else:
            reader = geojson.loads(data)
        
        total = list()
        # JSON files from Central use value as the keyword, whereas
        # GeoJSON uses features for the same thing.
        if 'value' in reader:
            data = reader['value']
        elif 'features' in reader:
            data = reader['features']
        for row in data:
            # log.info(f"ROW: {row}")
            tags = dict()
            if 'geometry' in row:
                tags['geometry'] = row['geometry']
            if 'properties' in row:
                indata = row['properties'] # A GeoJson formatted file
            else:
                indata = row    # A JOSM file from ODK Central
            for keyword, value in indata.items():
                # There's many extraneous fields in the input file which we don't need.
                base = keyword.lower()
                if (
                    base is None
                    or base in self.ignore
                    or value is None
                    or len(value) == 0
                ):
                    continue
                if keyword is None or len(keyword) == 0:
                    continue
                if type(value) == str:
                    items = self.convertEntry(keyword, value)
                    tags.update(items)
                else:
                    alltags = self.getAllTags(value)
                    for k, v in alltags.items():
                        # if k in self.ignore:
                        #     continue
                        if v:
                            items = dict()
                            if type(v) == dict:
                                v1 = alltags[k]
                                if len(v1) > 1:
                                    log.warning("Got more than 1 result! {v1}")
                                    v2 = v1[v1.keys()]
                                    items = self.convertEntry(k, v2)
                            else:
                                items = self.convertEntry(k, v)
                                #    for entry in items:
                                #        for k, v in entry.items():
                                #            tags[k] = v
                                #    else:
                                #tags[k1] = v1
                        tags.update(items)
            total.append(tags)
        return total

    def createEntry(self,
                    entry: dict
                    ):
        """Create the feature data structure"""
        # print(line)
        feature = dict()
        attrs = dict()
        tags = dict()
        priv = dict()
        refs = list()

        # log.debug("Creating entry")
        # First convert the tag to the approved OSM equivalent
        for key, value in entry.items():
            attributes = (
                "id",
                "timestamp",
                "lat",
                "lon",
                "uid",
                "user",
                "version",
                "action",
            )
            # When using existing OSM data, there's a special geometry field.
            # Otherwise use the GPS coordinates where you are.
            lat = None
            lon = None
            if type(value) == float:
                continue
            if key == "geometry":
                # The GeoJson file has the geometry field. Usually it's a dict
                # but on occasion it's a string instead, so turn it into a list
                # log.debug(f"FIXME: {value} {type(value)}")
                if type(value) == str:
                    if value[0] ==  '[':
                        coords = eval(value)
                        lat = coords[1]
                        lon = coords[0]
                    else:
                        coords = value.split(' ')
                        lat = coords[0]
                        lon = coords[1]
                    # log.debug(f"VALUE STRING: {coords}")
                elif type(value) == geojson.geometry.Point:
                    lat = value['coordinates'][1]
                    lon = value['coordinates'][0]
                    # log.debug(f"VALUE POINT: {lat}/{lon}")
                elif type(value) == list:
                    lat = float(value[1])
                    lon = float(value[0])
                attrs["lat"] = lat
                attrs["lon"] = lon
                # log.debug(f"ATTRS: {attrs}")

            if key is not None and len(key) > 0 and key in attributes:
                attrs[key] = value
                log.debug("Adding attribute %s with value %s" % (key, value))
            else:
                if key in self.multiple:
                    for item in value:
                        if key in item:
                            for entry in item[key].split():
                                vals = self.getValues(key)
                                if entry in vals:
                                    if vals[entry].find("="):
                                        tmp = vals[entry].split("=")
                                        tags[tmp[0]] = tmp[1]
                                else:
                                    tags[entry] = "yes"
                    continue

                if value is not None and value != "no" and value != "unknown":
                    if key == "track" or key == "geoline":
                        refs.append(tag)
                        log.debug("Adding reference %s" % tag)
                    elif len(str(value)) > 0:
                        if self.privateData(key):
                            priv[key] = value
                        else:
                            tags[key] = value
            if len(tags) > 0:
                if 'geometry' in tags:
                    del tags['geometry']
                feature["attrs"] = attrs
                feature["tags"] = tags
            if len(refs) > 0:
                feature["refs"] = refs
            if len(priv) > 0:
                feature["private"] = priv

        return feature


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="convert JSON from ODK Central to OSM XML"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-y", "--yaml", help="Alternate YAML file")
    parser.add_argument("-x", "--xlsfile", help="Source XLSFile")
    parser.add_argument(
        "-i", "--infile", help="The input file downloaded from ODK Central"
    )
    args = parser.parse_args()
    
    # if verbose, dump to the terminal.
    if args.verbose is not None:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        root.addHandler(ch)

    if args.yaml:
        jsonin = JsonDump(args.yaml)
    else:
        jsonin = JsonDump()
    # jsonin.parseXLS(args.xlsfile)

    # Modify the input file name for the 2 output files, which will get written
    # to the current directory.
    infile = Path(args.infile)
    base = os.path.splitext(infile.name)[0]
    osmoutfile = f"{base}-out.osm"
    jsonin.createOSM(osmoutfile)

    jsonoutfile = f"{base}-out.geojson"
    jsonin.createGeoJson(jsonoutfile)

    log.debug("Parsing JSON file %r" % args.infile)
    data = jsonin.parse(infile.as_posix())
    # This OSM XML file only has OSM appropriate tags and values
    for entry in data:
        feature = jsonin.createEntry(entry)
        # Sometimes bad entries, usually from debugging XForm design, sneak in
        if len(feature) == 0:
            continue
        if len(feature) > 0:
            if "lat" not in feature["attrs"]:
                if 'geometry' in feature['tags']:
                    if type(feature['tags']['geometry']) == str:
                        coords = list(feature['tags']['geometry'])
                        # del feature['tags']['geometry']
                    else:
                        coords = feature['tags']['geometry']['coordinates']
                        # del feature['tags']['geometry']
                    feature['attrs'] = {'lat': coords[1], 'lon': coords[0]}
                else:
                    log.warning("Bad record! %r" % feature)
                    continue
            jsonin.writeOSM(feature)
            # This GeoJson file has all the data values
            jsonin.writeGeoJson(feature)

    jsonin.finishOSM()
    jsonin.finishGeoJson()
    log.info("Wrote OSM XML file: %r" % osmoutfile)
    log.info("Wrote GeoJson file: %r" % jsonoutfile)
